#!/usr/bin/env python
########################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
########################################################################################
#
# General functions library for gLite Installer
#
# Authors: Robert Harakaly <robert.harakaly@cern.ch>
#          Alberto Di Meglio <alberto.di.meglio@cern.ch>
#          Diana Bosio <diana.bosio@cern.ch>
#          Joachim Flammer <Joachim.Flammer@cern.ch>
#          Guillermo Diez-Andino <Guillermo.Sancho@cern.ch>        
#
# $Id: gLiteInstallerLib.py,v 1.150.2.23.4.5.2.79.2.9.2.3 2007/11/06 14:29:25 dshiyach Exp $
########################################################################################

import os, string, socket, re, pwd, grp, time, stat
import sys, posix
import getopt
import urllib
import httplib
import commands
from random import choice

from xml.dom.minidom import Node
from xml.dom.minidom import parseString
from Ft.Xml import InputSource
from Ft.Xml import XPath as xpath
from Ft.Xml.Lib.Print import Print, PrettyPrint
from Ft.Xml.Xslt import Processor
from Ft.Xml.Domlette import ConvertDocument
from Ft.Xml.Domlette import NonvalidatingReader

class ConfigParams:
    def __init__(self, params = {}):
        self.params = params
        self.keyList = self.params.keys()
        
    def __getitem__(self, name):
        value = self.params[name]
        #if not isinstance(value,(list, dict, gLib)):
        try:
            if value.find('$${') > -1:
                value = string.replace(value, '$${', '${')
            else:
                while value.find('${') > -1:
                    #print value
                    rname = value[value.find('${')+2:value.find('}')]
                    try:
                        value = value.replace('${%s}'% rname, self.params[rname])
                    except KeyError:
                        try:
                            value = value.replace('${%s}'% rname, os.environ[rname])
                        except KeyError:
                            gLib().printErrorMessage("Referenced parameter '%s' not find in configuration files nor in environment" % rname)
                            break
        except AttributeError:
            pass
            
        return value

    def __setitem__(self, name, value):
        self.params[name] = value

    def __iter__(self):
        self.iterPosition = 0
        return(self)

    def next(self):
        if self.iterPosition >= len(self.keyList):
            raise StopIteration
        x = self.keyList[self.iterPosition]
        self.iterPosition += 1
        return x

    def has_key(self, name):
        return self.params.has_key(name)

    def update(self, name):
        return self.params.update(name)

    def keys(self):
        return self.params.keys()

    def values(self):
        return self.params.values()

    def copy(self):
        return self.params.copy()

    def __len__(self):
        return len(self.params)

    def __contains__(self, key):
        return key in self.params

##
#    class StringWriter
#        Utility class implementing the 'Writer' interface on top of 
#        the string

class StringWriter:
    def __init__(self):
        self.string = ""
    def write(self, string):
        self.string = self.string + string
    def getString(self):
        return self.string

class MyURLopener(urllib.FancyURLopener):
    def http_error_default(self, url, fp, errcode, errmsg, headers):
        void = fp.read()
        fp.close()
        raise IOError, ('http error', errcode, errmsg, headers)
urllib._urlopener = MyURLopener()

__all__ = ["GetoptError", "error", "getopt"]

global me
global params
global verbose

#    Code for nicer output of the PrettyPrint
#import xml.dom.ext
from xml.dom import XML_NAMESPACE, XMLNS_NAMESPACE, DOMException
##from xml.dom.ext import Printer

##
#    Following class/function extends the functionality of PrettyPrint 
#
#    As an output it produces text representation of DOM tree with similar
#    formatting as in the gLite configuration files
#
#    Usage
#        from UltraPrint import UltraPrettyPrint
#        from xml.dom.minidom import parse
#        dom = parse('test.xml')
#        UltraPrettyPrint(dom)

##class UltraPrintVisitor(Print.PrintVisitor):
##    import xml.dom.ext
##    length = 80
##    def visitAttribute(self, node):
##        if node.namespaceURI == XMLNS_NAMESPACE:
##            # Skip namespace declarations
##            return
##        self._write('\n' + self._indent * (self._depth + 1) + node.name)
##        value = node.value
##       if value or not self._html:
##            text = Printer.TranslateCdata(value, self.encoding)
##            text, delimiter = Printer.TranslateCdataAttr(text)
##            while text.find('  ') > -1:
##                text = text.replace('  ', ' ')     
##            n_text = ''
##            while len(text) > self.length:
##                s_pos = text.rfind(' ', 0, self.length + 1)
##                if s_pos == -1:
##                        s_pos = text.find(' ')     
##                        if s_pos == -1:
##                            s_pos = len(text)
##                n_text = n_text + text[0:s_pos] + '\n' + self._indent * (self._depth + 2)
##                text = text[s_pos + 1:len(text)]
##            text = n_text + text
##            self.stream.write("=%s%s%s" % (delimiter, text, delimiter))
##        return

##
#    Improved PrettyPrint functionality
#
#    @param root root element of the XML tree to print
#    @param stream stream to output to (should implement the Writer interface
#    @param encoding XML encoding
#    @param indent PrettyPrint indent
#    @param preserveElements PrettyPrint preserveElements

from Ft.Xml.Lib.Print import PrettyPrint as UltraPrettyPrint

##def UltraPrettyPrint(root, stream=sys.stdout, encoding='UTF-8', indent='  ', 
##                preserveElements=None):
##    if not hasattr(root, "nodeType"):
##        return
##    from xml.dom.ext import Printer
##    nss_hints = '' #SeekNss(root)
##    preserveElements = preserveElements or []
##    owner_doc = root.ownerDocument or root
##    if hasattr(owner_doc, 'getElementsByName'):
##        #We don't want to insert any whitespace into HTML inline elements
##        preserveElements = preserveElements + HTML_4_TRANSITIONAL_INLINE
##    visitor = UltraPrintVisitor(stream, encoding, indent, 
##                                   preserveElements, nss_hints)
##    Printer.PrintWalker(visitor, root).run()
##    stream.write('\n')
##    return
#########################################################################
#        End Ultra Pretty Print
#########################################################################


######################################
#    Simplified parser of the classAds files
######################################
class ClassAdsParser:
    def __init__(self, name=""):
        self.fileName = name
        
    def getConfig(self):
        self.params={}
        infile = open(self.fileName, "r")
        line = infile.readline()
        self.nodes(infile, "")
        infile.close()
        return self.params
                
    def nodes(self, infile, base):
        if self.verbose:
            print "\n"
        leave = 1
        while leave:
            line = infile.readline()
            if line.find(']') > -1:
                leave = 0
            else:
                if line.find('=') > -1:
                    if base == '' :
                        _base = line.split('=')[0].strip()
                    else:
                        _base = base + "." + line.split('=')[0].strip()
                    if line.find('[') > -1:
                        self.nodes(infile, _base)
                    else:
                        name = _base
                        value = line.split('=')[1].strip()
                        if value.find('"') > -1:
                            value = value[1:value.rfind('"')]
                        else:
                            value = value[0:value.rfind(';')]
                        self.params[name] = value
                        if self.verbose:
                            print "%s = %s" % (name, value)
                    
    def writeParams(self, fileName):

        outfile = open(fileName, "w")
        
        outfile.write("[\n")
        for section in self.params:
            if section != "/":
                outfile.write("    %s = [\n" % section)
            for param in self.params[section]:
                if isinstance(self.params[section][param], list):
                    outfile.write("        %s = {\n" % param)
                    for idx in xrange(len(self.params[section][param])):
                        if isinstance(self.params[section][param][idx], list):
                            outfile.write("            {")
                            for idx2 in xrange(len(self.params[section][param][idx])):
                                outfile.write(self.params[section][param][idx][idx2])
                                if idx2 < len(self.params[section][param][idx])-1:
                                    outfile.write(", ")
                            outfile.write("}")
                            if idx < len(self.params[section][param])-1:
                                outfile.write(",")
                            outfile.write("\n")
                        else:
                            outfile.write("            %s" % self.params[section][param][idx])
                            if idx < len(self.params[section][param])-1:
                                outfile.write(",")
                            outfile.write("\n")
                    outfile.write("        };\n")
                else:
                    separator = ' '
                    if self.params[section][param].strip().startswith('-') or \
                            self.params[section][param].strip().startswith('+'):
                        separator = ''
                    outfile.write("        %s =%s%s;\n" %(param, separator, self.params[section][param].strip()))
            if section != "/":
                outfile.write("    ];\n")
        outfile.write("]\n")

        outfile.close()
        
        return 0
                    
    def readParams(self, fileName):
        self.params = {}

        infile = open(fileName, "r")
        
        lines = []
        start = 0
        for line in infile:
            if line.strip() == "]":
                break;
            if start and line.strip() != "":
                lines.append(line)
                if line.find("];") != -1:
                    self.params.update(self.parseSection(lines))
                    lines = []
            if line.strip() == "[":
                start = 1
            
        infile.close()

        return 0
                    
    def parseSection(self, lines):
        
        section = {}
        start = 0
        isArray = 0
        for line in lines:
            if line.find("];") != -1:
                break
            if start and line.strip() != "" and not line.strip().startswith('//'):
                if isArray:
                    if line.find("};") != -1:
                        if len(lineArray) > 0:
                            section[sectionName][param] = self.parseArray(lineArray)
                        else:
                            section[sectionName][param] = []
                        isArray = 0
                    else:
                        lineArray.append(line)
                else:
                    param = line.split("=")[0].strip()
                    value = line.split("=")[1].strip()
                    value = value[0:value.rfind(';')]
                    section[sectionName][param] = value
                if line.strip().endswith("{"):
                    lineArray = []
                    param = line.split("=")[0].strip()
                    isArray = 1
            if line.find("[") != -1:
                start = 1
                sectionName = line.split("=")[0].strip()
                section[sectionName] = {}

        return section
        
    def parseArray(self, lineArray):
    
        valueArray = []
        for line in lineArray:
            if line.strip() != "":
                value = line.strip()
                value = value[0:value.rfind(',')]
                valueArray.append(value)
                
        return valueArray
    
    def updateValue(self, section, param, value, item=""):
        
        if item == "":
            self.params[section][param] = value
        else:
            self.params[section][param][value] = item

        return 0
        
    def addValue(self, section, param, value):
        
        self.params[section][param] = value

        return 0
        
    def deleteValue(self, section, param):
        
        try:
            del self.params[section][param]
        except:
            pass

        return 0
        
    def addSection(self, section):
        
        self.params[section] = {}

        return 0
        
##
#    Utility class for manipulation of the environment files

class ExportFile:
    def __init__(self, params, name=''):
        self.glib = gLib()
        self.glib.set_params(params)
        self.params = params
        if name == '':
            self.fileName = self.glib.getInstallerExportFile()
        else:
            self.fileName = name

    def source(self):
        #print "Sourcing %s file" % self.fileName
        f = open(self.fileName, 'r')
        for line in f.readlines():
            name_value = line.split(' ')[1]
            name = name_value.split('=')[0].strip()
            value = name_value.split('=')[1].strip()
            if value.find("$"+name) > -1:
                if os.environ.has_key(name):
                    value = string.replace(value, "$%s" % name, os.environ[name])
            os.environ[name] = value
            self.params[name] = value
            #print "%s = %s" % (name, os.environ[name])
        f.close()
            
    ##
    #     export
    #
    #     export environment variable to the external file
    #
    #     @param name name of environment variable to be set
    #     @param key parameter key to be used
    #     @fallBackValue  fallback value to be used if key does not exist
    def export(self, name, key='', fallBackValue=''):
        if not os.environ.has_key(name):
            if not self.params.has_key(name):
                if (not self.params.has_key(key)) and fallBackValue=='':
                    print "WARNING: no value/fallBackValue defined"
                    return 1
                if not self.params.has_key(key):
                    value = fallBackValue
                else:
                    value = self.params[key]
                # situation exactly as in the standard export method
                if not self.params.has_key(name):
                    os.environ[name]=value
                    self.params[name]=value
                else:
                    self.printWarningMessage("Parameter %s already defined, exporting %s value" % (name, self.params[name]))
                    os.environ[name]=self.params[name]
            else:
                os.environ[name] = self.params[name]
        if not os.path.exists(self.fileName):        
            f = open(self.fileName, 'w')
            f.close()
        
        # open files
        fin = open(self.fileName, 'r')
        fout = open(self.fileName+"~", 'w')
        
        # new line to be added
        gen_line="export %s=%s\n" %(name, os.environ[name])
        skip = 0
        
        # look if line exists already
        for line in fin.readlines():
            if line.find("export %s=" % name) > -1:
                fout.write(gen_line)
                skip = 1
            else:
                fout.write(line)
        
        # add line if not yet existing
        if skip == 0: 
            fout.write(gen_line)
        fin.close()
        fout.close()
        
        # move file
        os.rename(self.fileName+"~", self.fileName)
        return 0

try:
    from Ft.Xml import XPath
    from Ft.Xml import XInclude
    from Ft.Xml.FtMiniDom.DomTree import Node
    from Ft.Xml.FtMiniDom.NonvalReader import PyexpatNonvalReader
    from Ft.Xml.XPath import Context
    import urlparse

##
#    Utility class for keeping compatibility with newer 4Suite versions
#    Used only with old 4Suite parsers. Fixes bugs in the xinclude functionality

    class gliteXMLReader:
        ##
        #    reads XML file from URI, then calls processing of xincludes
        #    @param uri uri to read XML from
        #    @param reader reader to use for XML reading
        #    @param visited internal list to define already visited includes
        #    @param baseUri base URI to be used in case of relative paths
        #    @param uriResolver meaning to be determined
        def ProcessIncludesFromUri(self, uri, reader=None, visited=None, baseUri='', resolver=None):

            if not(urlparse.urlparse(uri)[0] == ''):
                baseUri = uri[0:uri.rfind('/')+1]
            reader = reader or PyexpatNonvalReader()
#            if resolver:
#                reader.uriResolver = resolver
            src = InputSource.InputSource(None, baseUri, processIncludes=0, stripElements=None, resolver=resolver)
            src = src.resolve(uri, '', '')
            node = reader.fromSrc(src)
            return self.ProcessIncludesFromNode(node, reader, visited, baseUri, resolver)

        ##
        #    processes xincludes from the XML tree
        #    @param node XML node containing the XML document
        #    @param reader reader to use for XML reading
        #    @param visited internal list to define already visited includes
        #    @param baseUri base URI to be used in case of relative paths
        def ProcessIncludesFromNode(self, node, reader=None, visited=None, baseUri='', resolver=None):

            nss={'xi': XInclude.XINCLUDE_NAMESPACE}
            reader = reader or PyexpatNonvalReader()
            con = Context.Context(node, processorNss=nss)
            g_includeExpr = XPath.Compile("//xi:include")
            nodes = g_includeExpr.evaluate(con)
            visited = visited or []
            for n in nodes:
                self._ProcessNode(n, visited, reader, baseUri, resolver)
            return node

        ##
        #    private method for processing of the xinclude tag
        #    @param node XML node containing the XML document
        #    @param reader reader to use for XML reading
        #    @param visited internal list to define already visited includes
        #    @param baseUri base URI to be used in case of relative paths
        def _ProcessNode(self, node, visited, reader, baseUri, resolver):

            if (node.namespaceURI, node.localName) != (XInclude.XINCLUDE_NAMESPACE, 'include'):
                return
            href = node.getAttributeNS(None, 'href')
            if not href:
                raise XInclude.XIncludeException(XInclude.XIncludeException.XINCLUDE_MISSING_HREF)
            if href in visited:
                raise XInclude.XIncludeException(XInclude.XIncludeException.CIRCULAR_INCLUDE_ERROR)
            parse = node.getAttributeNS(None, 'parse')
            xpointer = node.getAttributeNS(None, 'xpointer')
            if not parse or parse=='xml':
                replaceNode = self.ProcessIncludesFromUri(href, visited=visited,
                                              reader=reader, baseUri=baseUri, resolver=resolver)
                if replaceNode.nodeType == Node.DOCUMENT_NODE:
                    replaceNode = replaceNode.documentElement
                visited.append(href)
            else:
                if os.path.exists(href):
                    st = open(href)
                else:
                    st = urllib.open(href)
                replaceNode = node.ownerDocument.createTextNode(st.read())
                st.close()

            if xpointer != '':
                parent = node.parentNode
                for n in xpath.Evaluate("%s/*" % xpointer, replaceNode):
                    parent.appendChild(n)
                parent.removeChild(node)
            else:
                node.parentNode.replaceChild(replaceNode, node)
            visited.remove(href)
            return

except ImportError:
    print "Import error"
    pass

##
#    gLib class implementing gLite installer functionality
class gLib:

    def __init__(self):
        self.verbose = 0
        self.debug = 0
        h_name_file = os.popen('/bin/hostname -f')
        self.fq_hostname = h_name_file.read().strip()
        h_name_file.close()
        if not os.environ.has_key('HOSTNAME'):
            os.environ['HOSTNAME'] = self.fq_hostname
        self.hostname = os.environ["HOSTNAME"].split('.')[0]
        self.instances = {}
        self.parameters = {}
        self.params = ConfigParams()
        self.sets = {}
        self.nodeRoot = None
        
    def system(self, command):
        return os.system("source %s; %s" % (self.getInstallerExportFile(), command))
    
    def registerInitd(self):
        if not os.path.exists('/etc/init.d/gLite'):
            os.system("cp %s/gLite /etc/init.d/gLite" % self.getScriptPath())
            os.system("/bin/chmod 0755 /etc/init.d/gLite")
            os.system("chkconfig --levels 345 gLite on")
            # run checkconfig
        else:
            #  check if init.d script is enabled in the chkconfig
            if os.system("chkconfig  --list | grep gLite | grep [35]:on > /dev/null"):
                os.system("chkconfig --levels 345 gLite on")


        
    def set_params(self, params):
        self.params=params
        
        
    ##
    #    print parameters
    #    @param params dictionary of the parameters
    #    @return void
    def print_params(self, params):
        print "#-------------------------------------------------------------------"
        print "# Configuration parameters:"
        print "#-------------------------------------------------------------------\n"

        list = params.keys()
        list.sort()
        
        for x in list:
            self.printInfoMessage('%s = %s ' % (x, params[x]))
    
        print "#-------------------------------------------------------------------\n"

    def loadEnvFile(self, name, params):
        if os.path.exists(name) & self.verbose:
            print "File %s found" %name
        if os.path.exists(name):
            my_file =  open(name, 'r')
            for line in my_file:
                # srtrip strips the line of the new character
                # split creates a key pair separated at =
                # lower converts to lower case
                print type(line.rstrip().split("=")[0])
                print line.rstrip().split("=")[0]
                if params.has_key(line.rstrip().split("=")[0]):
                    print "Overwriting the " + line.rstrip().split("=")[0] + " variable"
                params[line.rstrip().split("=")[0]] = line.rstrip().split("=")[1]
    
    #-----------------------------------------------------------------------------
    # Read environment variables from default /etc/glite.conf and $HOME/.glite/glite.conf
    #-----------------------------------------------------------------------------

    def getEnvironment(self, params):
        self.loadEnvFile("/opt/glite/etc/glite.conf", params)
        self.loadEnvFile("/etc/glite/glite.conf", params)
        self.loadEnvFile("os.environ['HOME']/.glite/glite.conf", params)

    def getInstallerExportFile(self):
        if self.params.has_key('installer.export.filename'):
            installer_export_filename = self.params['installer.export.filename']
        else:
            installer_export_filename = '/etc/glite/profile.d/glite_setenv.sh'

        if not os.path.exists(os.path.dirname(installer_export_filename)):
            os.makedirs(os.path.dirname(installer_export_filename), 0755)
        if not os.path.exists(installer_export_filename):
            os.system("/bin/touch %s" % installer_export_filename)

        return installer_export_filename
    
    def source(self):
        exportFile = ExportFile(self.params, self.getInstallerExportFile())
        exportFile.source()


    def export(self, name, value=''):
        # exportFile = ExportFile(self.getInstallerExportFile())
        # exportFile.export(name,value)
        
        if not self.params.has_key(name):
            os.environ[name]=value
            self.params[name]=value
        else:
            os.environ[name]=self.params[name]
        
        self.unprotectEnvFile()
        
        #########################################
        #
        #    existing exports are overwritten
        #
        #########################################
        
        if not os.path.exists(self.getInstallerExportFile()):
            f = open(self.getInstallerExportFile(), 'w')
            f.close()
            
        if os.environ[name].strip().find(' ') > 0:
            gen_line = 'export %s="%s"\n' %(name, os.environ[name])
        else:
            gen_line = 'export %s=%s\n' %(name, os.environ[name])

        file = open(self.getInstallerExportFile(), 'r')
        for line in file.readlines():
            if line.find("export %s=" % name) != -1:
                break
        else:
            file.close()
            file = open(self.getInstallerExportFile(), 'a')
            file.write(gen_line)

        file.close()

        fin = open(self.getInstallerExportFile(), 'r')
        fout = open(self.getInstallerExportFile() + '.1', 'w')
        for line in fin.readlines():
            if line.find("export %s=" % name) == -1:
                fout.write(line)
            else:
                fout.write(gen_line)
        
        fin.close()
        fout.close()
        
        os.rename(self.getInstallerExportFile() + ".1", self.getInstallerExportFile())

        #########################################
        #
        #    generating the csh script.
        #    existing exports are overwritten
        #
        #########################################
        
        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())

        if not os.path.exists(cshFileName):
            f = open(cshFileName, 'w')
            f.close()

        if os.environ[name].strip().find(' ') > 0:
            gen_line = 'setenv %s "%s"\n' %(name, os.environ[name])
        else:
            gen_line = 'setenv %s %s\n' %(name, os.environ[name])

        file = open(cshFileName, 'r')
        for line in file.readlines():
            if line.find("setenv %s " % name) != -1:
                break
        else:
            file.close()
            file = open(cshFileName, 'a')
            file.write(gen_line)

        file.close()

        fin = open(cshFileName, 'r')
        fout = open(cshFileName + '.1', 'w')
        for line in fin.readlines():
            if line.find("setenv %s " % name) == -1:
                fout.write(line)
            else:
                fout.write(gen_line)
        
        fin.close()
        fout.close()
        
        os.rename(cshFileName + ".1", cshFileName)

        return 0

    #================================================================================
    # exportToLdSoConf: export a lib path to the ld.so.conf file and runs ldconfig
    #    Parameters:
    #        path: the path to export
    #================================================================================
    def exportToLdSoConf(self, path):
        
        if os.geteuid() == 0:
        
            if not os.path.exists('/etc/ld.so.conf'):
                os.system('/bin/touch /etc/ld.so.conf')
    
            file = open('/etc/ld.so.conf', 'r')
            for line in file:
                if line.find(path.strip()) != -1:
                    break
            else:
                file.close()
                file = open('/etc/ld.so.conf', 'a')
                file.write(path + "\n")
                
            file.close()
            
            os.system('ldconfig')
            
        else:
            
            self.printWarningMessage("Please remember to modify the file /etc/ld.so.conf by adding the line:")
            self.printWarningMessage(path)

        return 0

    def protectEnvFile(self):
        if not os.path.exists(self.getInstallerExportFile()):
            f = open(self.getInstallerExportFile(), 'w')
            f.close()
            
        protectionline = 'if [ -z "${GLITE_ENV_SET}" ]; then\n\n'
        infile = open(self.getInstallerExportFile(), 'r')
        outfile = open(self.getInstallerExportFile() + '.1', 'w')
        outfile.write(protectionline)
        for line in infile.readlines():
            if line.find("export") != -1 and line.find("GLITE_ENV_SET") == -1:
                if line.find('export GLITE_LOCATION=') != -1:
                    outfile.write('if [ -z "$GLITE_LOCATION" ]; then export GLITE_LOCATION=%s; fi\n' % os.environ['GLITE_LOCATION'])
                else:
                    outfile.write(line)
        outfile.write("\nexport GLITE_ENV_SET=true\n")
        outfile.write("\nfi\n")

        infile.close()
        outfile.close()
        os.rename(self.getInstallerExportFile() + ".1", self.getInstallerExportFile())

        #########################################
        #
        #    generating the csh script.
        #########################################

        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())

        if not os.path.exists(cshFileName):
            f = open(cshFileName, 'w')
            f.close()

        protectionline = "if ( ! $?GLITE_ENV_SET ) then\n\n"
        infile = open(cshFileName, 'r')
        outfile = open(cshFileName + '.1', 'w')
        outfile.write(protectionline)
        for line in infile.readlines():
            if line.find("set") != -1 and line.find("GLITE_ENV_SET") == -1:
                if line.find('setenv GLITE_LOCATION ') != -1:
                    outfile.write('if ( ! $?GLITE_LOCATION ) setenv GLITE_LOCATION %s \n' % os.environ['GLITE_LOCATION'])
                else:
                    outfile.write(line)
        outfile.write("\nsetenv GLITE_ENV_SET true\n")
        outfile.write("\nendif\n")

        infile.close()
        outfile.close()
        os.rename(cshFileName + ".1", cshFileName)

        return 0
        
    def unprotectEnvFile(self):

        if not os.path.exists(self.getInstallerExportFile()):
            f = open(self.getInstallerExportFile(), 'w')
            f.close()
            
        protectionline = 'if [ -z "${GLITE_ENV_SET}" ]; then'
        infile = open(self.getInstallerExportFile(), 'r')
        outfile = open(self.getInstallerExportFile() + '.1', 'w')
        for line in infile.readlines():
            if line.find(protectionline) == -1 and line.find('export GLITE_ENV_SET=') == -1 and line.strip() != 'fi':
                    outfile.write(line)

        infile.close()
        outfile.close()
        os.rename(self.getInstallerExportFile() + ".1", self.getInstallerExportFile())

        #########################################
        #
        #    generating the csh script.
        #########################################

        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())

        if not os.path.exists(cshFileName):
            f = open(cshFileName, 'w')
            f.close()

        protectionline = "if ( ! $?GLITE_ENV_SET ) then"
        infile = open(cshFileName, 'r')
        outfile = open(cshFileName + '.1', 'w')
        for line in infile.readlines():
            if line.find(protectionline) == -1 and line.find('setenv GLITE_ENV_SET') == -1 and line.strip() != 'endif':
                    outfile.write(line)

        infile.close()
        outfile.close()
        os.rename(cshFileName + ".1", cshFileName)

        return 0
        
    ##################################################
    # unexport
    #
    # unexport environment variable by removing it from setenv files
    #
    # params name  name of environment variable to be unset
    #
    ##################################################
    def unexport(self, name, value=''):

        # ------------------------------------------------------------        
        # Remove entry from setenv.sh file
        # ------------------------------------------------------------            

        # check if setenv file exists already
        if os.path.exists(self.getInstallerExportFile()):
                        
            os.rename(self.getInstallerExportFile(), self.getInstallerExportFile() + ".1")
            fin = open(self.getInstallerExportFile() + ".1", 'r')
            fout = open(self.getInstallerExportFile(), 'w')
        
            for line in fin.readlines():
                # write line for different parameters
                if line.find("export %s=" % name) == -1:
                    fout.write(line)
                        
            fin.close()
            fout.close()
            
            os.remove(self.getInstallerExportFile() + ".1")
        
        # ------------------------------------------------------------        
        # Remove entry from setenv.csh file
        # ------------------------------------------------------------            
        
        # replace xxx.sh by xxx.csh to get .csh setenv file
        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())

        # check if setenv file exists already
        if os.path.exists(cshFileName):

            os.rename(cshFileName, cshFileName + ".1")
            fin = open(cshFileName + ".1", 'r')
            fout = open(cshFileName, 'w')

            for line in fin.readlines():
                # write line for different parameters
                if line.find("export %s=" % name) == -1:
                    fout.write(line)
                        
            fin.close()
            fout.close()

            os.remove(cshFileName + ".1")
    
        return 0  
            
    def checkSetEnvFile(self, envFile):
        
        if not os.access(envFile, os.W_OK):
            try:
                self.check_dir(os.path.dirname(envFile))
                os.system("touch %s" % os.path.basename(envFile))
                print "Creating the environment file %s ...            " % envFile, 
                self.printOkMessage()
            except:
                print "Creating the environment file %s ...            " % envFile, 
                self.printFailedMessage()
                self.printWarningMessage("Changing to default file %s ...       " % "%s/.glite/glite_setenv.sh" % os.environ['HOME'])
                self.check_dir("%s/.glite" % os.environ['HOME'], 0755, os.getuid(), os.getgid())
                envFile = "%s/.glite/glite_setenv.sh" % os.environ['HOME']
                
        return envFile
        
    def setUserEnv(self, username=""):
        
        # Check parameter
        if not self.params.has_key('modify.user.env'):
            self.params['modify.user.env'] = 'true'
            
        # Do not modify if user has disabled this functionality
        if self.params['modify.user.env'] == 'false':
            return 0
            
        if username == "":
            username = pwd.getpwuid(os.getuid())[0]
        
        # do not change root environment   
        if self.params.has_key('modify.root.env') and \
            self.params['modify.root.env'] == 'false' and \
            username == 'root':
                return 0
            
        homePrefix = pwd.getpwnam(username)[5]
        
        if not os.path.exists(homePrefix):
            print "The %s user home directory was listed in the /etc/passwd file, but it could not be found" % username
            print "Please check your configuration."
            return 1

        # Properly adjust .bash_profile for non-interactive use
        if not os.path.exists("%s/.bash_profile" % homePrefix):
            os.system("/bin/touch %s/.bash_profile" % homePrefix)
        if self.verbose:
            print "Check .bash_profile content"
        f = open("%s/.bash_profile" % homePrefix, 'r')
        for line in f.xreadlines():
            if (line.find("BASH_ENV=~/.bashrc") != -1) or (line.find("BASH_ENV=$HOME/.bashrc") != -1):
                f.close()
                break
        else:
            f.close()
            f = open("%s/.bash_profile" % homePrefix, 'a')
            f.write("\nexport BASH_ENV=~/.bashrc")
            f.close()

        f = open("%s/.bash_profile" % homePrefix, 'r')
        for line in f.xreadlines():
            if line.find(". ~/.bashrc") != -1:
                f.close()
                break
        else:
            f.close()
            f = open("%s/.bash_profile" % homePrefix, 'a')
            f.write("\n. ~/.bashrc\n")
            f.close()

        # Source glite_setenv from the account .bashrc file
        if not os.path.exists("%s/.bashrc" % homePrefix):
            os.system("/bin/touch %s/.bashrc" % homePrefix)
        if self.verbose:
            print "Check .bashrc content"
        f = open("%s/.bashrc" % homePrefix, 'r')
        for line in f.xreadlines():
            if line.find(". %s" % self.getInstallerExportFile()) != -1:
                f.close()
                break
        else:
            f.close()
            f = open("%s/.bashrc" % homePrefix, 'a')
            f.write("\n# gLite environment")
            f.write("\n. %s\n" % self.getInstallerExportFile())
            f.close()
        
        # Source glite_setenv from the account .cshrc file
        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())
        if not os.path.exists("%s/.cshrc" % homePrefix):
            os.system("/bin/touch %s/.cshrc" % homePrefix)
            os.system("/bin/chown %s:%s %s/.cshrc" % (username, pwd.getpwnam(username)[3], homePrefix))
            os.system("/bin/chmod 0644 %s/.cshrc" % homePrefix)
        if self.verbose:
            print "Check .cshrc content"
        f = open("%s/.cshrc" % homePrefix, 'r')
        for line in f.xreadlines():
            if line.find("source %s" % cshFileName) != -1:
                f.close()
                break
        else:
            f.close()
            f = open("%s/.cshrc" % homePrefix, 'a')
            f.write("\n# gLite environment")
            f.write("\nsource %s\n" % cshFileName)
            f.close()
        
        # Source glite_setenv from the account .tcshrc file
        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())
        if not os.path.exists("%s/.tcshrc" % homePrefix):
            os.system("/bin/touch %s/.tcshrc" % homePrefix)
            os.system("/bin/chown %s:%s %s/.tcshrc" % (username, pwd.getpwnam(username)[3], homePrefix))
            os.system("/bin/chmod 0644 %s/.tcshrc" % homePrefix)
        if self.verbose:
            print "Check .tcshrc content"
        f = open("%s/.tcshrc" % homePrefix, 'r')
        for line in f.xreadlines():
            if line.find("source %s" % cshFileName) != -1:
                f.close()
                break
        else:
            f.close()
            f = open("%s/.tcshrc" % homePrefix, 'a')
            f.write("\n# gLite environment")
            f.write("\nsource %s\n" % cshFileName)
            f.close()
            
        return 0
        
    def check_existing(self, name, message = ""):
        if not os.path.exists(name):
            if message == "":
                self.printErrorMessage("File/dir %s does not exist" % name)
            else:
                self.printErrorMessage(message)
            return 1
        return 0
            
    # check directory permissions and ownership
    def check_dir(self, name, perms=0, uid=os.getuid(), gid=os.getgid(), fix=1):
        
        name = self.expandPath(name)
        
        if os.path.isfile(name):
            return 0
        
        if not os.path.exists(name):
            if fix != 1:
                self.printErrorMessage("Directory %s does not exist" % name)
                return 1
            else:
                os.makedirs(name)
                self.printInfoMessage("Created directory %s " % name)

        if self.check_owner(name, pwd.getpwuid(uid)[0], grp.getgrgid(gid)[0], fix):
            return 1
            
        if not (perms == 0):
            if self.check_perms(name, perms, fix):
                return 1

        return 0
            
    # check file permissions and ownership
    def check_file(self, name, perms=0, uid=os.getuid(), gid=os.getgid(), fix=1):
        
        name = self.expandPath(name)
        
        if not os.path.exists(name):
            return 0
            
        if self.check_owner(name, pwd.getpwuid(uid)[0], grp.getgrgid(gid)[0], fix):
            return 1
            
        if not (perms == 0):
            if self.check_perms(name, perms, fix):
                return 1

        return 0
            
    def chown_r(self, name, uid, gid):
        
        name = self.expandPath(name)
        uid = int(uid)
        gid = int(gid)
        os.system("/bin/chown -R %d:%d %s" % (uid, gid, name))
        return 0
                
    # Obsolete will be deleted
    def check_dir_perms(self, name, perms):
        
        name = self.expandPath(name)
        self.check_dir(name, perms)
        return 0

    #================================================================================
    # check_perms: check directory/file permissions and optionally fix them
    #
    #    Parameters:
    #        name: the directory/file path to check
    #        perms: the octal permission mask to check
    #        fix: if fix = 1, permissions will be fixed if wrong, otherwise an error message
    #            is output
    #================================================================================
    def check_perms(self, name, perms, fix=1):

        name = self.expandPath(name)
        st = os.stat(name)
        mode = oct(st[stat.ST_MODE] & 0777)
        if self.verbose:
            print "Mode of file or directory %s is " % name, mode
            
        if mode != oct(perms):
            if fix != 1:
                self.printErrorMessage("Permissions on file or directory %s are incorrect. It should be %s" % (name, oct(perms)))
                return 1
            else:
                if self.verbose:
                    print "Changing mode of file or directory %s to %s" % (name, oct(perms))
                os.system("/bin/chmod %s %s" % (oct(perms), name))
            
        return 0

    #================================================================================
    # check_owner: check directory/file ownership and optionally fix them
    #
    #    Parameters:
    #        name: the directory/file path to check
    #        user: the username to check (not the uid)
    #        group: the group name to check (not the gid)
    #        fix: if fix = 1, ownership will be fixed if wrong, otherwise an error message
    #            is output
    #================================================================================
    def check_owner(self, name, user, group, fix=1):

        name = self.expandPath(name)

        st = os.stat(name)
        try:
            uid = st[stat.ST_UID]
            cur_user = pwd.getpwuid(uid)[0]
            gid = st[stat.ST_GID]
            cur_group = grp.getgrgid(gid)[0]
        except Exception, msg:
            self.printErrorMessage("Could not retrieve user and group name from given ids. The error is:")
            self.printErrorMessage(msg)
            return 1
            
        if self.verbose:
            print "Owner of file or directory %s is %s:%s" % (name, cur_user, cur_group)
            
        wrong = 0
        if user != cur_user or group != cur_group:
            if fix != 1:
                self.printErrorMessage("Ownership on file or directory %s is incorrect. It should be %s:%s" % (name, user, group))
                wrong = 1
            else:
                if self.verbose:
                    print "Changing owner of file or directory %s to %s:%s" % (name, user, group)
                os.chown(name, pwd.getpwnam(user)[2], grp.getgrnam(group)[2])
            
        if wrong and fix != 1:
            return 1
            
        return 0

    ##############################################################################
    #
    #    Add "path" into environment path
    #
    ##############################################################################
    
    def addEnvPath(self, name, value):

        if not os.path.exists(self.getInstallerExportFile()):
            f = open(self.getInstallerExportFile(), 'w')
            f.close()
                    
        pathlist = string.split(value, ':')
        pathlist.append('')
        for path in pathlist:
          if not path == '':
            if name == "MANPATH":
                gen_line = 'if [ -z "$%s" ]; then export %s=%s:; else export %s=%s::$%s; fi\n' \
                    % (name, name, path, name, path, name)
            else:
                gen_line = 'if [ -z "$%s" ]; then export %s=%s; else export %s=%s:$%s; fi\n' \
                    % (name, name, path, name, path, name)
            f = open(self.getInstallerExportFile(), 'r')
            for line in f.xreadlines():
                if line.find("=" + path + ":") != -1:
                    break
            else:
                f.close()
                f = open(self.getInstallerExportFile(), 'a')
                f.write(gen_line)

            f.close()

       ########################################################################
       #
       # Add path to tcsh environment.
       # path is added using the set notation rather than setenv
       #
       #######################################################################
    
        regexpr = re.compile('\.sh$')
        cshFileName = regexpr.sub('.csh', self.getInstallerExportFile())
        
        if not os.path.exists(cshFileName):
            f = open(cshFileName, 'w')
            f.close()
            
        fw = open(cshFileName + ".1", 'w')
        fw.write('if ( ! $?path ) set path= \n')
        if name.strip() != "PATH":
            fw.write('if ( ! $?%s ) setenv %s \n' % (name, name))
        fr = open(cshFileName, 'r')
        for line in fr.xreadlines():
            if line.find('if ( ! $?%s )' % name) == -1 and line.find('if ( ! $?path )') == -1:
                fw.write(line)
        fr.close()
        fw.close()
        os.rename(cshFileName + ".1", cshFileName)

        for path in pathlist:
            if ((name.strip() == "path") or (name.strip() == "PATH")):
                gen_line = 'set path=(%s $path)\n' % (path)
            else:
                gen_line = 'setenv %s %s:$%s\n' % (name, path, name)
            f = open(cshFileName, 'r')
            for line in f.xreadlines():
                if line.find(" " + path + ":") != -1 or line.find("=(" + path + " ") != -1:
                    f.close()
                    break
            else:
                f.close()
                f = open(cshFileName, 'a')
                f.write(gen_line)
                f.close()

        # Add current path
        if os.environ.has_key(name):
            for path in pathlist:
                if os.environ[name].find(path) == -1:
                    os.environ[name] = path + ":" + os.environ[name]
        else:
            os.environ[name] = value

    ###############################################################################
    #
    # Group and user management functions
    #
    ################################################################################
    #-------------------------------------------------------------------------------
    # Add a user to a group, if the group doesn't exist it is created
    # Return the user uid and gid if successful, -1,-1 if failed
    #-------------------------------------------------------------------------------
    def add_user(self, uname, group='', uid=0, gid=0):

        uid = int(uid)
        gid = int(gid)
        
        if group != '':
            gid = self.add_group(group, gid)
            if gid == 0:
                return -1, -1
        
        # Check if user exists
        try:
            user = pwd.getpwnam(uname)
            if (uid != 0) and (user[2] != uid):
                print "The user %s already exist with uid=%d and you asked for uid=%d" % (uname, pwd.getpwnam(uname)[2], uid)
                return -1, -1
            elif (group != '') and (user[3] != grp.getgrnam(group)[2]):
                print "The user %s already exist, adding to group %s" % (uname, group)
                os.system('/usr/sbin/usermod -G %s %s' % (group, uname))
            else:
                if self.verbose:
                    print "The user %s already exist" % uname
        except:
            # Check if a group with the same username exist
            try:
                gr = grp.getgrnam(uname)
                group = uname
            except:
                pass

            # User doesn't exist, create it add it to the specified group
            if group == '':
                if uid != 0:
                    os.system('/usr/sbin/useradd -u %s %s' % (uid, uname))
                else:
                    os.system('/usr/sbin/useradd %s' % uname)
            else:
                if uid != 0:
                    os.system('/usr/sbin/useradd -g %s -u %s %s' % (group, uid, uname))
                else:
                    os.system('/usr/sbin/useradd -g %s %s' % (group, uname))
            
        # Set user environment
        self.setUserEnv(uname)

        # Return the numeric uid,gid
        return self.get_user_info(uname)

            
    def get_user_info(self, name):
        # Return the numeric uid,gid
        try:
            uid = pwd.getpwnam(name)[2]
            gid = pwd.getpwnam(name)[3]
        except:
            uid = gid = -1

        return uid, gid
            
    def create_accounts(account_list):
        userlist = string.split(account_list, ',')
        for u in userlist:
            (user, group) = string.split(u, ':')
            self.add_user(user, group)
    
    def checkInteractiveLogin(self, username):
        errorcode = os.system('su - %s -c ""' % username)
        if errorcode:
            self.printErrorMessage("The account %s does not have interactive login enabled" % username)
            self.printErrorMessage("Please verify the /etc/security/limits.conf file before continuing")
            return 1
        return 0
        
    #-------------------------------------------------------------------------------
    # Add a group, if the group doesn't exist it is created
    # If a GID is specified the user is created with the given GID, unless a group
    # with the same name already exists
    #-------------------------------------------------------------------------------
    
    def add_group(self, gname, gid=0):
    
        gid = int(gid)
        
        # Check if group exists, if not create it
        try:
            gr = grp.getgrnam(gname)
            if (gid != 0):
                if gr[0] == gname and gr[2] == gid:
                    if self.verbose:
                        self.printInfoMessage("Group %s already exists with given GID (%s)" % (gname, gid))
                    return int(gid)
                if gr[0] == gname and gr[2] != gid:
                    self.printErrorMessage("Group %s already exists with differenr GID (%s). Cannot continue." % (gname, gid))
                    return 0
            else:
                if gr[0] == gname:
                    if self.verbose:
                        self.printInfoMessage("Group %s already exists" % gname)
                    return int(gr[2])
        except:
            if gid == 0:
                os.system('/usr/sbin/groupadd %s' % gname)
            else:
                os.system('/usr/sbin/groupadd -g %s %s' % (gid, gname))

        # Verify that group has been created
        try:
            gr = grp.getgrnam(gname)
            return int(gr[2])
        except:
            self.printErrorMessage("Failed to add group %s " % gname)
            return 0
    
    
    ###############################################################################a#
    #
    # Process management functions
    #
    ################################################################################
    
    #-------------------------------------------------------------------------------
    # getPID: get the PID of the first occurrence of a matching
    # running process or 0 if no match is found
    #-------------------------------------------------------------------------------
 
    def getPID(self, pname=''):
        if pname == '':
            print 'Process name argument is empty'
            return 0

        proclist = os.popen('/bin/ps axww | grep %s' % pname)
        proc = proclist.readlines()
        for item in proc:
            if (item.find('grep %s' % pname) == -1) and \
                        (item.find('%s' % pname) != -1) and \
                        (item.find('-config.py') == -1):
                rec = string.splitfields(string.strip(item), ' ')
                return int(rec[0])
        else:
            return 0   
    #-------------------------------------------------------------------------------
    # startProcess: runs a service and checks if it has been successfully started
    # process=process name     command=command to run the process
    # In case it started correctly returns 0, if not 1
    #-------------------------------------------------------------------------------
    def startProcess(self, process, command):
        pid = self.getPID(process)
        if pid != 0:
                print "The %s service is already running     " % process, 
                self.printOkMessage()
        else:
            os.system(command)
            pid = self.getPID(process)
            if pid != 0:
                print "The %s service was started    " % process, 
                self.printOkMessage()
                return 0
            else:
                print "Failed to start the %s service    " % process, 
                self.printFailedMessage()
                return 1
                    
        return 0
                                                                                                                                                             
    #-------------------------------------------------------------------------------
    # stopProcess: stops a service that is running. process = name of the process
    # id no command is given, the process is killed
    # 0 if everything goes OK, 1 on the contrary
    #-------------------------------------------------------------------------------
    def stopProcess(self, process, command=""):
        pid = self.getPID(process)
        if pid == 0:            
            print "The %s service is not running     " % process
            print "Stopping the %s service           " % process, 
            self.printOkMessage()
        else:
            if command != "":
                os.system(command)
            else:
                os.system("kill -9 %d" % pid)
            time.sleep(2)
            pid = self.getPID(process)
            if pid == 0:
                print "Stopping the %s service           " % process, 
                self.printOkMessage()
                return 0
            else:
                self.printWarningMessage("Failed to stop the %s service     " % process)
                print "Stopping the %s service " % process, 
                self.printFailedMessage()
                return 2
                
        return 0

    #-------------------------------------------------------------------------------
    # getScriptPath: get the path of the current script
    # Parameter: scriptName - Script file name
    #-------------------------------------------------------------------------------
    
    def getScriptPath(self):
        pathPrefix = os.path.split(sys.argv[0])[0]
        if pathPrefix == '':
            pathPrefix = '.'
        return pathPrefix

    ###############################################################################a#
    #
    # XML management functions
    #
    ################################################################################
    
    
    def file_add_tag(self, pfilename, text):
        file = open(filename, 'a')
        file.write("#<"+me+">\n")
        file.writelines(text)
        file.write("#</"+me+">\n")
        file.close()
       
    def file_delete_tag(self, pfilename):
        write = 1
        fromfile = open(filename, 'r')
        tofile = open(filename + ".new", "w")
        for line in fromfile:
            if (line == ("#<"+me+">\n")):
                write = 0
            if write:
                tofile.write(line)
            if (line == ("#</"+me+">\n")):
                write = 1
        tofile.close()
        fromfile.close()
        os.rename(filename + ".new", filename)
    
    def file_update_tag(self, pfilename, text):
        file_delete_tag(filename)
        file_add_tag(filename, text)
    
    
    ###############################################################################a#
    #
    # Certificate management functions
    #
    ################################################################################
    
    #-------------------------------------------------------------------------------
    # Check to see if host certs exist and that the permissions are correct
    #-------------------------------------------------------------------------------
    
    def check_certs(self, params, fix=1):

        error_level = 0
        
        #-------------------------------------------------------------------
        # Check host certificates exist
        #-------------------------------------------------------------------

        if not params.has_key('host.certificate.file'):
            hostcert = '/etc/grid-security/hostcert.pem'
            params['host.certificate.file'] = '/etc/grid-security/hostcert.pem'
        else:
            hostcert = params['host.certificate.file']
        if not params.has_key('host.key.file'):
            hostkey = '/etc/grid-security/hostkey.pem'
            params['host.key.file'] = '/etc/grid-security/hostkey.pem'
        else:
            hostkey = params['host.key.file']
    
        if self.check_existing(hostcert, "Host certificate not found at %s" % hostcert) != 0:
            error_level = 1
        else:
            if self.verbose:
                print "Host certificate %s found" % hostcert
            if self.check_perms(hostcert, 0644, 0):
                error_level = 2
            if self.check_owner(hostcert, 'root', 'root'):
                error_level = 2
        if self.check_existing(hostkey, "Host key not found at %s" % hostkey) != 0:
            error_level = 1
        else:
            if self.verbose:
                print "Host key %s found" % hostkey
            if self.check_perms(hostkey, 0400, 0):
                error_level = 2
            if self.check_owner(hostkey, 'root', 'root'):
                error_level = 2

        if fix and error_level == 2:
            self.check_file(hostcert, 0644, 0, 0)
            self.check_file(hostkey, 0400, 0, 0)
            self.printInfoMessage("Host key and host certificate permissions and ownership have been fixed")
            error_level = 0
        
        print "Checking if host key and host certificate are correctly installed    ", 
        if error_level:
            self.printFailedMessage()
        else:
            self.printOkMessage()
        
        return error_level
    
    #-------------------------------------------------------------------------------
    # Get the subject of a given certificate
    # The default separator for DNs parts is ',' , any other character in sep
    # produces a '/' separator
    # Returns the certificate subject if successful
    # Raises AttributeError exception if failed
    #-------------------------------------------------------------------------------
        
    def get_certificate_subject(self, filename, sep=','):
        if sep == ',':
            nameopt = "-nameopt sep_comma_plus_space"
        else:
            nameopt = ""
        subject = os.popen("openssl x509 -in %s -noout -subject %s" % (filename, nameopt))
        for line in subject:
            if line.find("subject") != -1:
                return line.split("subject=")[1].strip()
                break
        else:
            raise AttributeError, "Unable to get certificate subject"
            

    #################################################################
    # checkConfigurationFile
    #
    # check format of configuration file
    #
    # parameter location   location of configuration file (can be also remote)
    #           verbose    default 1
    #
    # return 0 ok
    #        1 error
    #
    #################################################################
    def checkConfigurationFile(self, location, verbose=1):
        
        status_output = commands.getstatusoutput("/usr/bin/xmllint %s" % location)
        status = status_output[0]/256
        output = status_output[1]
    
        if status == 0:
            return 0
        else:
            if verbose == 1:
                print output
            return 1

    ##
    #     Load all configuration values for the service
    #
    #    @param filename filename or list of the filenames to read to get all of the configuration
    #    @param dictionary containing the read parameters
    #
    #    @return error code (0/1)
    def loadConfiguration(self, filename, params):
        
        # set default parameters
        self.setGlobalDefaultParameters(params)
        
        siteUrl = ''
        # Check if site.config.url was given on the command line
        site_config_url = ""
        if params.has_key('site.config.url'):
            if params['site.config.url'] != "":
                site_config_url = params['site.config.url']
        
        if not isinstance(filename, list):
            path = filename[0:string.rfind(filename, "/")]
        else:
            path = filename[0][0:string.rfind(filename[0], "/")]
            
        # load service configuration 
        if not isinstance(filename, list):
            filename = [ filename ]

        configUrl = [self.getConfigurationLocalFileURL("%s/glite-global.cfg.xml" % path)]
        for name in filename:
                configUrl.append(self.getConfigurationLocalFileURL(name))

        if self.loadConfigurationFiles(configUrl, params) > 0:
            self.printErrorMessage("An error has occurred while parsing the configuration files")
            return 1
        
        # Get environment from all possible configuration files ...
        self.getEnvironment(params)
    
        self.params = params
        
        return 0
        
    ##
    #     Load configuration values from a URL
    #
    #    @param url URL to read from
    #    @params dictionary of the parameters to use
    #
    #    @return returns: 0/1 error code of the loadConfigurationFile call
    def loadConfigurationURL(self, url, params):
        if self.verbose:
            print "Loading configuration from URL %s" % url
        
        try:
            file = urllib.urlopen(url)
        except Exception, msg:
            self.printErrorMessage("The site configuration URL is not valid")
            self.printErrorMessage("The error is: %s" % msg)
            return 1
        
        errorcode = self.loadConfigurationFile(url, params)
        
        return errorcode


    ##
    #     Load configuration values from a local XML file
    #
    #    @param name file name to read from
    #    @params dictionary of the parameters to use
    #
    #    @return returns: 0/1 error code of the loadConfigurationFile call
    def loadConfigurationLocalFile(self, name, params):
        if self.verbose:
            print "Loading configuration from local file %s" % name
        errorcode = self.loadConfigurationFile("file://" + os.path.abspath(name), params)
        if errorcode:
            self.printErrorMessage("Could not open the configuration file %s " % name)
            self.printErrorMessage("or the file is not a valid xml file.")
        
        return errorcode

    def getConfigurationLocalFileURL(self, name):
        return "file://" + os.path.abspath(name)
    
    ##
    #     private method to load configuration values by using the URI
    #
    #    @param uri URI to read from
    #    @params dictionary of the parameters to use
    #
    #    @return returns: 0/1 error code of the loadConfigurationFile call
    def loadConfigurationFiles(self, fileURIs, params):

        stringWriter = StringWriter()
        siteStringWriter = StringWriter()
        scriptPrefix = os.path.abspath(self.getScriptPath())

#       parserMethod = NonvalidatingReader.parseUri
        parserMethod = gliteXMLReader().ProcessIncludesFromUri

        self.proc1 = Processor.Processor()
        self.proc2 = Processor.Processor()

        volist_sty = InputSource.DefaultFactory.fromUri("file://%s/volist.xsl" % scriptPrefix)
        sty = InputSource.DefaultFactory.fromUri("file://%s/instance_iterate.xsl" % scriptPrefix)
        self.proc1.appendStylesheet(volist_sty)
        self.proc2.appendStylesheet(sty)

        from Ft.Lib.Uri import BaseUriResolver
        from Ft.Lib import UriException
        
        resolver = BaseUriResolver()
        resolver.supportedSchemes.append("https")
        resolver.supportedSchemes.append("file")

        globalStringWriter = StringWriter()
        try:
            if self.verbose:
                print "Loading " + fileURIs[0]
            UltraPrettyPrint(parserMethod(fileURIs[0], resolver=resolver), globalStringWriter)
        except:
            self.printErrorMessage("%s cannot be loaded" % fileURIs[0])
            return 1

        globalDOM = NonvalidatingReader.parseString (globalStringWriter.getString(), "dummy_uri")

        siteURI = xpath.Evaluate('//site.config.url', globalDOM)[0].getAttributeNS(None, 'value').strip()

        if siteURI =='':
            siteDom = NonvalidatingReader.parseString ("<config/>", "dummy_uri")
        else:
            siteDom = NonvalidatingReader.parseString ("<siteconfig/>", "dummy_uri")

        self.updateDom(globalDOM.documentElement, siteDom.documentElement, "/*/*")

        fileURIs.remove(fileURIs[0])

        if not siteURI == '':

            if self.verbose:
                print "Loading configuration from %s" % siteURI
            siteStringWriter = StringWriter()
            try:
                if self.verbose:
                    print "Loading " + siteURI
                UltraPrettyPrint(parserMethod(siteURI, resolver=resolver), siteStringWriter)
            except IOError:
                self.printErrorMessage("siteconfig URL %s is not valid" % siteURI.strip())
                return 1
            except UriException, e :
                self.printErrorMessage(e.message)
                return 1

            siteDomObject = NonvalidatingReader.parseString (siteStringWriter.getString(), "dummy_uri")
            self.updateDom(siteDomObject.documentElement, siteDom.documentElement, "/*/*")

        self.localDom = NonvalidatingReader.parseString ("<config/>", "dummy_uri").documentElement
        
        for file in fileURIs:

            try:
                confWriter = StringWriter()
                UltraPrettyPrint(parserMethod(file, resolver=resolver), confWriter)
                if self.verbose:
                    print "Loading configuration from %s" % file
                fileDom = NonvalidatingReader.parseString (confWriter.getString(), "dummy_uri").documentElement
                self.updateDom(fileDom, self.localDom, "/*/*")

            except:
                pass

        if siteURI == '':
            self.updateDom(self.localDom, siteDom.documentElement, "/*/*")
        else:
            self.updateDom(self.localDom, siteDom.documentElement, "/*/vo")

        stringWriter = StringWriter()
        UltraPrettyPrint(self.localDom, stringWriter)               

        xslResult = self.applyXSL(stringWriter)

        self.localDom = NonvalidatingReader.parseString (xslResult, "dummy_uri")
 
        stringWriter = StringWriter()
        UltraPrettyPrint(siteDom, stringWriter)

        result = self.applyXSL(stringWriter)

        doc = NonvalidatingReader.parseString (result, "dummy_uri")


        ##########################################################
        #    Printing of the resulting DOM tree after the transformations
        ##########################################################
        if self.debug:
            print "Siteconfig"
            print result
            print "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
            print "LocalFiles"
            UltraPrettyPrint(self.localDom)
            print "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"

        ##########################################
        #    Parse the configuration file structure
        #    Makes actions to read site configuration
        ##########################################
        rootNodes = doc.childNodes
        i = 0
        while rootNodes[i].nodeType != 1:
            i = i+1
        rootNode = rootNodes[i]
        rootName = rootNode.nodeName
        try:
            #########################################
            #    Siteconfig structure
            #########################################
            if rootName == 'siteconfig':

                hostFound = 0

                for node in rootNode.childNodes:
                    #################################
                    #    Take site globals
                    #################################
                    if node.nodeType == 1 and node.nodeName == 'parameters':
                        params.update(self.getXMLParams(node))
                    #################################
                    #    Take the concrete node config
                    #    equivalent to the standard config structure
                    #################################
                    if node.nodeType == 1 and node.nodeName == 'node':
                        hostlist = node.getAttributeNS(None, 'name').split(',')
                        if self.fq_hostname in hostlist:
                            hostFound = 1
                            params.update(self.getConfig(node)) 
                            self.nodeRoot = node;  
                            self.params.update(self.updateNode(params))

                    ##############################
                    #       Set section
                    ##############################
                    if node.nodeType == 1 and node.nodeName == "set":
                        #print "Found set"
                        self.sets[node.attributes['name'].nodeValue] = node
                        #self.sets[node.getAttributeNS('','name')] = node
                    ##############################
                    #       Instance section
                    ##############################
                    if node.nodeType == 1 and node.nodeName == "instance":
                        service = node.getAttributeNS(None, 'service')
                        if service == "":
                            service = 'default'
           
                        nodeName = node.getAttributeNS(None, 'name')
                        if (nodeName.find("changeme") != -1):
                            raise ValueError, "The instance name %s is not properly defined" % nodeName
                        instanceName="%s:%s" %(service, nodeName) 
                        if self.instances.has_key(instanceName):
                            self.updateParams(node, self.instances[instanceName])
                        else:
                            self.instances[instanceName] = node
                        if self.verbose:
                            print("Instance %s of service %s set" % (instanceName.split(':')[1], instanceName.split(':')[0]))
                            
                if not hostFound:
                    self.nodeRoot = rootNode
                    self.params.update(self.updateNode(params))
                    self.printWarningMessage("No matching computer hostname for %s could be found in the site configuration file" % self.fq_hostname) 
                    self.printWarningMessage("If this is not expected, please check the node tags in the file for any error or missing entry") 

            ##########################################
            #    Standard config structure
            ##########################################
            if rootName == 'config' or rootName == 'global':
                params.update(self.getConfig(rootNode))
                self.nodeRoot = rootNode
            self.params.update(params)
        except ValueError, msg :
            self.printErrorMessage(msg)
            return 1
        return 0
            
            
    ##
    #    applyXSL
    #        Calls the XSL scripts
    #    @param stringWriter Instance of the StringWriter class containing the DOM tree
    #    @return string representation of XML document after the transformatios
            
    def applyXSL(self, stringWriter) :

        from Ft.Xml import InputSource
        phase1 = self.proc1.run(InputSource.DefaultFactory.fromString(stringWriter.getString(), "dummy_uri"))
        result = self.proc2.run(InputSource.DefaultFactory.fromString(phase1, "dummy_uri"))
        return result
        
    ##
    #    Parses the standard (local config and node config) config structure
    #
    #    @param rootNode root node of the XML subtree to parse
    #
    #    @exception ValueError 
    #
    #    @return Dictionary with parsed parameters
    def getConfig(self, rootNode):
        params = {}
        nodes = rootNode.childNodes
        for node in nodes:
            if node.nodeType == 1:
        
                name = node.nodeName
                if name == "parameters":
                    ##############################
                    #    Standard parameters section
                    #    Main parameters section
                    ##############################
                    service = node.getAttributeNS(None, 'service')
                    if service == "":
                        # Standard parameters section without service attribute
                        params.update(self.getXMLParams(node))
                    else:
                        # service dependent parameters
                        self.parameters[service] = node
                    
                if name == "instance":
                    #################################
                    #    Instance section
                    #################################
                    #try:
                        #service = node.attributes['service'].nodeValue
                    service = node.getAttributeNS(None, 'service')
                    if service == "":
                    #except:
                        service = 'default'
           
                    #instanceName="%s:%s" %(service, node.attributes['name'].nodeValue)
                    nodeName = node.getAttributeNS(None, 'name')
                    if (nodeName.find("changeme") != -1):
                        raise ValueError, "The instance name %s is not properly defined" % nodeName
                    instanceName="%s:%s" %(service, nodeName) 
                    if self.instances.has_key(instanceName):
                        self.updateParams(node, self.instances[instanceName])
                    else:
                        self.instances[instanceName] = node
                    if self.verbose:
                        print("Instance %s of service %s set" % (instanceName.split(':')[1], instanceName.split(':')[0]))

                if name == "set":
                    ##############################
                    #    Set section
                    ##############################
                    #print "Found set"
                    #self.sets[node.attributes['name'].nodeValue] = node
                    setName = node.getAttributeNS(None, 'name')
                    if (setName.find("changeme") != -1):
                        raise ValueError, "The set name %s is not properly defined" % setName
                    self.sets[setName] = node
        
        self.params.update(self.updateNode(params))            
        return params             
        
    ##
    #    Update parameters and instances from local files
    #
    def updateNode(self, params):
        #    Update instances from local file
        localInstanceList = xpath.Evaluate('//instance', self.localDom)
        for instance in localInstanceList:
            instanceName = "%s:%s" %(instance.getAttributeNS(None, 'service'), instance.getAttributeNS(None, 'name'))
            if instanceName in self.instances.keys():
                #update instance
                self.updateParams(instance, self.instances[instanceName])
            else:
                #add instance
                self.instances[instanceName] = instance
                
        #    Update parameters from local file
        parametersNodes = xpath.Evaluate('/*/parameters', self.localDom)
        if len(parametersNodes) > 0:
            params.update(self.getXMLParams(parametersNodes[0]))
        
        return params
            
    ##
    #    Merge two DOM subtrees with overwriting of the parameter values
    #
    #    @param src source DOM node to be merged into the dest
    #    @param dest destination DOM node containing the result
    #    
    #    @return void
    def updateParams(self, src, dest):
        srcParamNameList = self.getNodeNameList(xpath.Evaluate('parameters[not(@service)]/*', src))
        destParamNameList = self.getNodeNameList(xpath.Evaluate('parameters[not(@service)]/*', dest))
        for node in srcParamNameList:
            if node in destParamNameList:
                srcNode = xpath.Evaluate('parameters[not(@service)]/%s' % node, src)[0]
                destNode = xpath.Evaluate('parameters[not(@service)]/%s' % node, dest)[0]
                parent = destNode.parentNode
                parent.removeChild(destNode)
                parent.appendChild(srcNode)    
            else:
                parent = xpath.Evaluate('parameters[not(@service)]', dest)[0]
                parent.appendChild(xpath.Evaluate('parameters[not(@service)]/%s' % node, src)[0].cloneNode(1))
        
    def updateDom(self, src, dest, expression):
        srcContainerNames = self.getContainerNameList(xpath.Evaluate(expression, src))
        dstContainerNames = self.getContainerNameList(xpath.Evaluate(expression, dest))
        for container in srcContainerNames.keys():
            if container in dstContainerNames.keys():
                #    Update existing container
                if srcContainerNames[container].nodeName == 'parameters':
                    self.updateParams(srcContainerNames[container].parentNode, dstContainerNames[container].parentNode)
                else:
                    self.updateParams(srcContainerNames[container], dstContainerNames[container])
            else:
                #    copy container from source to destination
                node = srcContainerNames[container]
                dest.appendChild(node.cloneNode(1))



    def getContainerNameList(self, nodes):
        list = {}
        for node in nodes:
            name = "%s:%s:%s" % (node.nodeName, node.getAttributeNS(None, "name"), node.getAttributeNS(None, "service"))
            list[name] = node
        return list

    ##
    #    private method to list the tag names in the NodeList
    #    @param nodeList NodeList to work on
    #
    #    @return list of the tag names
    def getNodeNameList(self, nodeList):
        list = []
        for node in nodeList:
            list.append(node.nodeName)
        return list

    ##
    #    Lists all defined instances for a given service
    #        
    #    @param service service name for which the instances should be listed
    #    
    #    @return list of instances defined for a given service 
    def getInstances(self, service='default'):
        list={}
        for key in self.instances.keys():
            if key.find(service) == 0:
                list[key] = self.instances[key]
        return list

    ##
    #    get the list of services known to the configuration
    #    
    #    @return list of services used
    def getServices(self):
        list = []
        for key in self.instances.keys():
            service = key.split(':')[0]
            if service not in list:
                list.append(service)
        return list

    ##
    #    Utility method to print all instance names to the screen
    def showInstances(self):
        for instanceName in self.instances.keys():
           print instanceName

    ##
    #    Toplevel method to filter XML nodes to inspect and to update parameter dictionaries
    #
    #    @param parametersNode node corresponding to the <parameters> tag
    #
    #    @exeption ValueError if user defined parameter is not provided
    #
    #    @return Dictionary of the parameters
    def getXMLParams(self, parametersNode):
        conf = {}
        for node in parametersNode.childNodes:
            if node.nodeType == Node.ELEMENT_NODE:
                conf.update(self.getXMLParam(node, conf))

        return conf
        
    ##
    #    private method which checks if the node contains "name' attribute
    #
    #    @param node element node to inspect
    #
    #    @return 0/1 
    def hasNameAttribute(self, node):
        if node.getAttributeNS(None, 'name') == '':
            return 0
        return 1

    ##
    #    Analyses the XML structure to get the parameters
    #    It also replaces $parameter by parameters value
    #
    #    @param parametersNode node corresponding to the <parameters> tag
    #    @param conf internal temporary dictionary to store parameters
    #
    #    @exeption ValueError if user defined parameter is not provided
    #
    #    @return Dictionary of the parameters            
    def getXMLParam(self, node, conf = {}):
        name = node.nodeName
        if node.hasChildNodes() and self.hasNameAttribute(node):
            subConf = {}
            conf[name] = subConf
            if (node.getAttributeNS(None, 'name').find("changeme") != -1):
                raise ValueError, "The parameter %s is not properly defined" % node.getAttributeNS(None, 'name')
            subConf['name'] = node.getAttributeNS(None, 'name')
            for subnode in node.childNodes:
                if subnode.nodeType == 1:
                        subConf[node.nodeName] = self.getXMLParams(node)
 
        if node.hasChildNodes() and not self.hasNameAttribute(node):
            ############################
            # Array values:
            #    Only arrays has childnodes 
            #    without attributes
            ############################
            #print "found array %s" % name
            if self.isSimpleArray(node):
                conf[name] = self.setArray(node)
            else:
                conf[name] = self.setTagArray(node)

        if not node.hasChildNodes():    
            ############################
            #    Regular configuration parameters
            ############################
            #    print node.nodeName
            #        value = node.attributes['value'].nodeValue
            value = node.getAttributeNS(None, 'value')

            if value.find("changeme") != -1:
                raise ValueError, "The user-defined parameter %s is not defined" \
                    % name
            conf[name] = value
#            if (value.find('$') > -1) and (value.find('$${') == -1):
#                ###############################
#                #    Replace values from already defined parameters
#                ################################
#                self.expandParams(conf,name,conf)
#                self.expandParams(conf,name,self.params)
#                self.expandParams(conf,name,os.environ)
#            elif (value.find('$${') != -1):
#                conf[name] = string.replace(value, '$${', '${')
                
        return conf
        
    ##
    #    Parameter expansion: $param or ${param}
    #    result is stored in the params dictionary
    #
    #    @param conf
    #    @param name
    #    @param params
    #
    #    @return void
#    def expandParams(self,conf, name, params):
        
#        for key in params.keys():
#            if (not conf[name].find("$") == conf[name].find("${")):
#                self.printWarningMessage("Ambiguous parameter replacement without {}. Please use ${}")
#                self.printWarningMessage(conf[name])

#            if (conf[name].find('$' + key) > -1):
#                conf[name] = string.replace(conf[name], '$' + key, params[key])
#            elif (conf[name].find('${'+key+'}') > -1):
#                conf[name] = string.replace(conf[name], '${' + key + '}', params[key])
#    
#        return 0

    ####################################################
    #    Expands env vars in name
    #    Returns: expanded path
    ####################################################
    def expandPath(self, name):
        
        for key in os.environ.keys():
            if (name.find('$' + key) > -1):
                name = string.replace(name, '$' + key, os.environ[key])
            elif (name.find('${' + key + '}') > -1):
                name = string.replace(name, '${' + key + '}', os.environ[key])
    
        return name

    #####################################################
    #     Enable definition of the array values
    #
    #    <parameters>
    #        <arrayName>
    #            <value>First line</value>
    #            <value>Second line</value>
    #        </arrayName>
    #     ....
    #
    #     Raises: ValueError in case of missing user defined parameter
    #
    #     Returns: array of values
    ######################################################
    
    def setArray(self, array, name=''):
        if self.params.has_key(name):
            ret = self.params[name] 
        else:
            ret = []
        for node in array.childNodes:
            if node.nodeType == node.ELEMENT_NODE:
                for item in node.childNodes:
                    if (item.nodeType == item.TEXT_NODE):
                        if item.data.find("changeme") != -1:
                            if not self.params.has_key(name):
                                raise ValueError, "The user-defined parameter %s is not defined" % name
                        else:
                            ret.append(item.data)
                    break
                else:
                    # This is an empty string node value
                    ret.append("")
                    
        return ret

    def setTagArray(self, array):
        ret = {}
        for item in array.childNodes:
            if (item.nodeType == item.ELEMENT_NODE):
                #ret[item.attributes['value'].nodeValue]=self.getXMLParams(item)
                if(item.getAttributeNS(None, 'value').find("changeme") != -1):
                    raise ValueError, "The user-defined parameter %s is not defined" % item.getAttributeNS(None, 'value')
                ret[item.getAttributeNS(None, 'value')] = self.getXMLParams(item)
        return ret

    #####################################################
    #    Name (service:name) defines an unique instance
    #
    #    Raises:
    #        ValueError exception from getXMLParams
    #        LookupError exception if the given instance name is not found
    #    This has to be trapped by caller
    #
    #    Returns: Dictionary with global and instance specific parameters
    #####################################################
    def getInstanceParams(self, name):

        cfg = ConfigParams()
        if name == 'default':
            cfg = self.params
        else:
            if not self.instances.has_key(name):
                raise LookupError, "Instance %s not defined in the cfg file" % name
            cfg = self.params.copy()
            # service dependent parameters section if exist ...
            if self.parameters.has_key(name.split(":")[0]):
                cfg.update(self.getXMLParams(self.parameters[name.split(":")[0]]))
            # read instance parameters
            for node in self.instances[name].childNodes:
                if node.nodeType == 1 and node.nodeName == "parameters":
                    cfg.update(self.getXMLParams(node))

        return ConfigParams(cfg)

    #######################################################
    #    getVoList
    #            Returns the VO names defined for given node
    #
    #    Returns: string array of names
    #    TODO: correct error handling with exception in case of the 
    #            problems instead of value '1'
    #######################################################
    def getVoList(self):
        list = []
        if self.nodeRoot == None:
            return 1
        nodeVOs = xpath.Evaluate("volist", self.nodeRoot)
        if len(nodeVOs) == 0:
            nodeVOs = xpath.Evaluate("/*/volist", self.nodeRoot)
        if len(nodeVOs) == 0:
            return list

        for node in nodeVOs:
            list.append(node.getAttributeNS(None, 'name'))
        return list

        
    def getVoParams(self, name):
        cfg = ConfigParams()
        parametersNode = xpath.Evaluate("//vo[@name='%s']" % name, self.nodeRoot)[0].getElementsByTagName("parameters")
        if len(parametersNode) > 0 :
            cfg = self.getXMLParams(parametersNode[0])
        return ConfigParams(cfg)
           
    def isSimpleArray(self, nodes):
        for node in nodes.childNodes:
            if node.nodeType == Node.ELEMENT_NODE and node.nodeName == 'value':
                return 1
        return 0
        
    def getSets(self):
        return self.sets.keys()
        
    def getSetParams(self, name):
        cfg = ConfigParams()
        for node in self.sets[name].childNodes:
            if node.nodeType == 1:
                cfg.update(self.getXMLParam(node))
        return ConfigParams(cfg)
        
    def getConfigPrefix():
        # Check if GLITE_LOCATION corresponds ot one from the configuration file path
        
        if os.geteuid() == 0:
            return self.params['GLITE_LOCATION']
        ##################
        #    Ordinary user
        ##################
        #    Check if I can write into the $GLITE_LOCATION
        if access(self.params['GLITE_LOCATION'], W_OK):
            return self.params['GLITE_LOCATION']
        else:
            return os.environ("HOME") + "/.glite"

    def registerService(self):
        self.registerInitd()
        name = os.path.split(sys.argv[0])[1]
        file = open("%s/etc/config/scripts/gLite.services" % self.params['GLITE_LOCATION'], "r")
        for line in file:
            if line.find(name) != -1 and not line.strip().startswith("#"):
                break
        else:
            file.close()
            file = open("%s/etc/config/scripts/gLite.services" % self.params['GLITE_LOCATION'], "a")
            file.write(name + "\n")

        file.close()

    #################################################################
    # printOkMessage
    #
    # function to print out "[  OK  ]" in color
    # 
    #################################################################
    def printOkMessage(self):
        print "[\033[32m  OK  \x1b[0m]"
    
    #################################################################
    # printFailedMessage
    #
    # function to print out "[FAILED]" in color
    # 
    #################################################################
    def printFailedMessage(self):
        print "[\033[31mFAILED\x1b[0m]"
        
    #################################################################
    # printErrorMessage
    #
    # function to print out "[ERROR]" in color
    # 
    #################################################################
    def printErrorMessage(self, message):
        print "[\033[31mERROR\x1b[0m] %s" % message
     
  
    #################################################################
    # printInfoMessage
    #
    # function to print out info message in color
    #
    # message error message
    # 
    #################################################################
    def printInfoMessage(self, message):
        print "[\033[32mINFO\x1b[0m]    %s" % message    
    
    #################################################################
    # printWarningMessage
    #
    # function to print out warning message in color
    #
    # message error message
    # 
    #################################################################
    def printWarningMessage(self, message):
        print "[\033[33mWARNING\x1b[0m] %s" % message
     
                
    #################################################################
    # checkJava
    #
    # function to check if java is actually installed in JAVA_HOME
    # 
    #################################################################
    def checkJava(self):

        print "Checking Java installation                               ", 
        
        if not os.environ.has_key('JAVA_HOME'):
            if not self.params.has_key('JAVA_HOME'):
                self.printFailedMessage()
                self.printErrorMessage("JAVA_HOME is not defined.")
                return 1
                
        if (os.environ['JAVA_HOME'] == "changeme"):
            self.printFailedMessage()
            self.printErrorMessage("JAVA_HOME not set correctly in the configuration files.")
            return 1
        
        if os.path.exists('%s/bin/java' % os.environ['JAVA_HOME']):
            if self.verbose:
                self.printOkMessage()
                print "Java found in %s/bin/java             " % os.environ['JAVA_HOME'], 
        else:            
            self.printFailedMessage()
            self.printErrorMessage("JAVA_HOME is not pointing to a valid java installation.")
            self.printErrorMessage("Please verify that the Java SDK or JRE are installed")
            self.printErrorMessage("and the JAVA_HOME parameter in the %s/etc/config/glite-global.cfg.xml file"  % os.environ['GLITE_LOCATION'])
            self.printErrorMessage("is correctly set")
            return 1
        
        self.printOkMessage()

        return 0

    #################################################################
    # checkJavaSDK
    #
    # function to check if the Java SDK is actually installed in JAVA_HOME
    # 
    #################################################################
    def checkJavaSDK(self):
        print "Checking Java installation                                ", 
        
        if not os.environ.has_key('JAVA_HOME'):
            if not self.params.has_key('JAVA_HOME'):
                self.printFailedMessage()
                self.printErrorMessage("JAVA_HOME is not defined.")
                return 1
                
        if (os.environ['JAVA_HOME'] == "changeme"):
            self.printFailedMessage()
            self.printErrorMessage("JAVA_HOME not set correctly in the configuration files.")
            return 1
        
        if os.path.exists('%s/bin/java' % os.environ['JAVA_HOME']) and \
            os.path.exists('%s/lib/tools.jar' % os.environ['JAVA_HOME']):
            if self.verbose:
                self.printOkMessage()
                print "Java SDK found in %s/bin/java             " % os.environ['JAVA_HOME'], 
        else:            
            self.printFailedMessage()
            self.printErrorMessage("JAVA_HOME is not pointing to a valid Java SDK installation.")
            self.printErrorMessage("Please verify that the Java SDK is installed")
            self.printErrorMessage("and the JAVA_HOME parameter in the %s/etc/config/glite-global.cfg.xml file"  % os.environ['GLITE_LOCATION'])
            self.printErrorMessage("is correctly set")
            return 1
        
        self.printOkMessage()

        return 0
    
    
    #################################################################
    # setGlobalDefaultParameters
    #
    # Sets the default parameters that are otherwise defined in the 
    # 'advanced' and 'system' section of the global configuration file
    # and the autocalculated values.
    #
    # Function should be called before parameters are loaded.
    #
    # params  params list of parameters
    #
    # return 0 ok
    #        1 error
    #
    #################################################################
    def setGlobalDefaultParameters(self, params):
        
        # -------------------------------------------------------------------------------        
        # advanced parameters
        # -------------------------------------------------------------------------------
        
        params['site.config.url'] = ""

        # -- gLite configuration --

        params['GLITE_LOCATION']     = "/opt/glite"
        params['GLITE_LOCATION_VAR'] = "/var/glite"
        params['GLITE_LOCATION_LOG'] = "/var/log/glite"
        params['GLITE_LOCATION_TMP'] = "/tmp/glite"

        # -- External applications and services --
        
        params['LCG_LOCATION']         = "/opt/lcg"
        params['EDG_LOCATION']         = "/opt/edg"
        params['GLOBUS_LOCATION']      = "/opt/globus"
        params['GPT_LOCATION']         = "/opt/gpt"
        params['CATALINA_HOME']        = "/var/lib/tomcat5"
        params['JAVA_HOME']            = "/usr/lib/jvm/java-1.5.0-sun-1.5.0.12/jre"
        params['tomcat.CATALINA_OPTS'] = "-Xmx${HALF_MEMORY_SIZE} -server -Dsun.net.client.defaultReadTimeout=240000"
            
        # -- Certificate settings --
        
        params['host.certificate.file'] = "/etc/grid-security/hostcert.pem"
        params['host.key.file']         = "/etc/grid-security/hostkey.pem"
        params['user.certificate.path'] = ".certs"
        params['host.gridmapfile']      = "/etc/grid-security/grid-mapfile"
        params['host.groupmapfile']     = "/etc/grid-security/groupmapfile"
        params['host.gridmap.dir']      = "/etc/grid-security/gridmapdir"
        
        params['host.groupmap.dir']     = "/etc/grid-security/groupmapdir"
        params['X509_VOMS_DIR']         = "/etc/grid-security/vomsdir"
        params['X509_CERT_DIR']         = "/etc/grid-security/certificates"
        params['ca.certificates.dir']   = "${X509_CERT_DIR}"
                    
        # -------------------------------------------------------------------------------
        # system parameters
        # -------------------------------------------------------------------------------

        params['installer.export.filename'] = "/etc/glite/profile.d/glite_setenv.sh"
        params['modify.user.env']           = "true"
        params['tomcat.user.name']          = "tomcat"
        params['tomcat.user.group']         = "tomcat"
            
        # -------------------------------------------------------------------------------
        # auto calculated values
        # -------------------------------------------------------------------------------
        
        # determine half memorysize (in MB)
        proclist = os.popen("free | grep ^Mem | awk '{print int($2/1024/2)}'")
        proc = proclist.readlines()
        for item in proc:
            halfMemorySize = item.strip()
        proclist.close()

        # half memory size
        params['HALF_MEMORY_SIZE'] = "%sM" % halfMemorySize
        
        return
    
    
    #################################################################
    # GenPasswd
    #
    # Generate a strong random password
    # 
    #################################################################
    def GenPasswd(self):
        
        chars = string.letters + string.digits
        
        newpasswd = ""
        
        for i in range(8):
            newpasswd = newpasswd + choice(chars)
        
        return newpasswd
        
    ################################################################
    # getopt
    #
    # Replaces the getopt.getopt function
    # 
    #################################################################   
    def getopt(self, args, shortopts, longopts = []):
        opts = []
        if type(longopts) == type(""):
            longopts = [longopts]
        else:
            longopts = list(longopts)

        while args and args[0].startswith('-') and args[0] != '-' and args[0] != "--":
            if args[0].startswith('--'):
                opts, args = getopt.do_longs(opts, args[0][2:], longopts, args[1:])
            else:
                opts, args = getopt.do_shorts(opts, args[0][1:], shortopts, args[1:])
        if not args:
            return opts, args
        else:
            raise getopt.GetoptError('option %s not valid' % args[0], args[0])
            
    ###################################
    #    addWatchdog
    ###################################
    def addWatchdog(self, service, command, result, startCommand, stopCommand):

        # create watchdog crontab
        crontab = open("/etc/cron.d/watchdog.cron", "w")
        crontab.write("PATH=/sbin:/bin:/usr/sbin:/usr/bin\nHOSTNAME=%s\nHOME=/\n\n\
*/2 * * * * root %s/etc/config/scripts/watchdog.py >> %s/watchdog.log  2>&1\n" \
            % (os.environ['HOSTNAME'], \
            os.environ['GLITE_LOCATION'], \
            os.environ['GLITE_LOCATION_LOG']))
        crontab.close()

        # Touch cron spool directory to cause reloading of the crontabs
        os.system("/bin/touch /var/spool/cron")
             
        from xml.dom.minidom import getDOMImplementation
    
        impl = getDOMImplementation()
    
        factory = impl.createDocument(None, "watchdog", None)
        root = factory.documentElement
    
        root.appendChild(self.createParameterNode(factory, "command", command))
        root.appendChild(self.createParameterNode(factory, "result-string", result))
        root.appendChild(self.createParameterNode(factory, "start-command", startCommand))
        root.appendChild(self.createParameterNode(factory, "stop-command", stopCommand))
    
        self.check_dir("%s/etc/watchdog.d" % os.environ['GLITE_LOCATION'])
        file = open("%s/etc/watchdog.d/%s.xml" % (os.environ['GLITE_LOCATION'], service), "w")
        xmlString = root.toxml()
        file.write(xmlString.replace("><", ">\n<"))
        file.close()

    #########################################
    # checkHTTPSconnection
    #
    # check a https via getXXX to check the status of a service
    #
    # params hostName    host name (e.g. lxb1440.cern.ch)
    #        endpoint    URL       (e.g. /EGEE/glite-data-metadata-catalog/services/MetaDataCatalog?method=getVersion)
    #        port        port to connect (default: '8443')
    #        keyFile     location of key file in pem format (default: '/etc/grid-security/hostkey.pem')
    #        certFile    location of cert file in pem format (default: '/etc/grid-security/hostcert.pem')
    #        
    # return status, code, description
    #        where status = 0: ok, 
    #              status = 1: failed
    #              code    return code from https request (200=ok,111=connection refused, 404=file not found ....)
    #              message message from https request
    #
    #########################################
    def checkHTTPSconnection(self, hostName, endpoint, port=8443, keyFile='/etc/grid-security/hostkey.pem', certFile='/etc/grid-security/hostcert.pem'):
        
        try:
            connection = httplib.HTTPSConnection(hostName, port, keyFile, certFile)
            connection.request("GET", endpoint)
            response = connection.getresponse()
        except socket.error, msg:
            return (1, msg[0], msg[1])

        # check if response was ok (200)    
        if response.status==200:
            return (0, response.status, response.reason)
        else:
            if response.status==404:
                return (1, response.status, "Endpoint '%s' not found." % response.reason)
            else:
                return (1, response.status, response.reason)            

    def createParameterNode(self, factory, name, value):
        node = factory.createElement(name)
        node.setAttribute("value", value)
        return node        

    def addLogrotate(self, service, logFile, command, logNumber = 5, period = "daily"):
        
        logrotate = open("/etc/logrotate.d/%s" % service, "w")
        logrotate.write("%s {\n" % logFile)
        logrotate.write("    rotate %s\n" %logNumber)
        logrotate.write("    %s\n" %period)
        logrotate.write("    postrotate\n        %s\n    endscript\n}" %command)
        logrotate.close()
        
        return 0
    
    #######################################
    # appendInstance
    #
    # Adds an instance to the instances list
    #
    # params 
    #        node     instance sub-dom-tree    
    #
    # TODO: Add merging of the instances if one already exist
    #######################################
    def appendInstance(self, node):
        name = node.getAttributeNS(None, 'name')
        service = node.getAttributeNS(None, 'service')
        self.instances["%s:%s" % (service, name)] = node
        
