#!/usr/bin/env python
#####################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
#####################################################################################
#
# Template script for configuring the Globus service for gLite
# Alberto Di Meglio <alberto.di.meglio@cern.ch>
# Version info: $Id$
#
#####################################################################################

import os,sys,string
sys.path.append(".")
from gLiteInstallerLib import gLib
from gLiteInstallerLib import ExportFile


class Globus:

    def __init__(self,params):
    	self.params = params
        self.glib=gLib()
        self.verbose = 0
#        self.globus_location = params['GLOBUS_LOCATION']
#        os.environ['GLOBUS_LOCATION'] = self.globus_location
        if not self.params.has_key('GLOBUS_LOCATION'):
            os.environ['GLOBUS_LOCATION'] = '/opt/globus'
            self.params['GLOBUS_LOCATION'] = os.environ['GLOBUS_LOCATION']
        if not self.params.has_key('GPT_LOCATION'):
            os.environ['GPT_LOCATION'] = '/opt/gpt'
            self.params['GPT_LOCATION'] = os.environ['GPT_LOCATION']
            
    def start(self):
        return 0

    def stop(self):
        return 0

    def configure(self):
        
        try:
            if os.path.exists(self.params['host.certificate.file']):
                subject = self.glib.get_certificate_subject(self.params['host.certificate.file'])
                subjectFields = subject.split(',')
            else:
                print "Could not find a valid host certificate at %s" % self.params['host.certificate.file']
                return 1
        except AttributeError, e:
            print e
            return 1
        
        # Create the gobus.conf configuration file
        if self.verbose:
            print "Creating the /etc/globus.conf file"
        config_file = ExportFile(self.params,"/etc/globus.conf")
        config_file.export("GLOBUS_LOCATION")
        config_file.export("GLOBUS_HOST_DN",'', '"' + subject + '"')
        config_file.export("GLOBUS_ORG_DN","", '"'+string.strip(subjectFields[0] + ',' + subjectFields[1])+'"')
        config_file.export("GRIDMAP","host.gridmap.file","/etc/grid-security/grid-mapfile")
        config_file.export("GRIDMAPDIR","host.gridmap.dir","/etc/grid-security/grid-mapdir")
        config_file.export("GSIWUFTPPORT",'',"2811")
        config_file.export("GSIWUFTPDLOG",'',"/var/glite/log")
        config_file.export('GLOBUS_FLAVOR_NAME','globus.flavor.name')
        config_file.export('X509_GATEKEEPER_CERT','',self.params['host.certificate.file'])
        config_file.export('X509_GATEKEEPER_KEY','',self.params['host.key.file'])
        config_file.export('X509_GSIWUFTPD_CERT','',self.params['host.certificate.file'])
        config_file.export('X509_GSIWUFTPD_KEY','',self.params['host.key.file'])
        config_file.export('X509_CERT_DIR','',self.params['ca.certificates.dir'])
 

#        f = open("/etc/globus.conf",'w')
#        f.write("export GLOBUS_LOCATION=%s\n" % self.params['GLOBUS_LOCATION'])
#        f.write("export GLOBUS_HOST_DN=%s\n" % subject)
#        f.write("export GLOBUS_ORG_DN=%s\n" % subjectFields[0] + ',' + subjectFields[1])
#        f.write("export GRIDMAP=%s\n" % self.params['host.gridmapfile'])
#        f.write("export GRIDMAPDIR=%s/gridmapdir\n" % os.path.dirname(self.params['host.gridmapfile']))
#        f.write("export GSIWUFTPPORT=%s\n" % self.params['gsiwuftpport'])
#        f.write("export GSIWUFTPDLOG=%s\n" % self.params['gsiwuftpdlog'])
#        f.write("export GLOBUS_FLAVOR_NAME=%s\n" % self.params['globusflavor'])
#        f.write("export X509_GATEKEEPER_CERT=%s\n" % self.params['x509_gatekeepercert'])
#        f.write("export X509_GATEKEEPER_KEY=%s\n" % self.params['x509_gatekeeperkey'])
#        f.write("export X509_GSIWUFTPD_CERT=%s\n" % self.params['x509_gsiwuftpdcert'])
#        f.write("export X509_GSIWUFTPD_KEY=%s\n" % self.params['x509_gsiwuftpdkey'])
#        f.write("export X509_CERT_DIR=%s\n" % self.params['x509_certdir'])
#        f.close()


        #self.postInstall()
        
        return 0

    def postInstall(self):

        # Run gpt-postinstall
        gpt_post = "%s/sbin/gpt-postinstall" % os.environ['GPT_LOCATION']
        if not os.path.exists(gpt_post):
            print "GPT doesn't seem to be correctly installed. Please verify"
            return 1
            
        if self.verbose:
            print "Running gpt-postintall..."
        os.system(gpt_post)
        
        # Run Globus setup
        globus_setup = "%s/setup" % os.environ['GLOBUS_LOCATION']
        if not os.path.exists(globus_setup):
            print "Globus doesn't seem to be correctly installed. Please verify"
            return 1

        gk_conf = "%s/etc/gatekeeper.conf" % os.environ['GLITE_LOCATION']
        if not os.path.exists(gk_conf):
            print "No gLite gatekeeper configuration file could be found in %s" % gk_conf
            return 1

        ggk_conf = "%s/etc/globus-gatekeeper.conf" % os.environ['GLOBUS_LOCATION']
        if os.path.exists(ggk_conf):
            os.remove(ggk_conf)

        os.system("%s/globus/setup-tmpdirs" % globus_setup)
        os.system("%s/globus/setup-globus-common" % globus_setup)
        os.symlink(gk_conf, ggk_conf)
        os.system("%s/globus/setup-globus-gram-job-manager" % globus_setup)
        os.system("%s/globus/setup-globus-job-manager-fork" % globus_setup)

        return 0

    def restart(self):
        return 0
