#!/usr/bin/env python
#####################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
#####################################################################################
#
# General security functions library for gLite Installers
# Authors: Robert Harakaly <robert.harakaly@cern.ch>
#        Alberto Di Meglio <alberto.di.meglio@cern.ch>
#        Diana Bosio <diana.bosio@cern.ch>
#        Joachim Flammer <Joachim.Flammer@cern.ch>
#        Guillermo Diez-Andino <Guillermo.Sancho@cern.ch>        
#
# $Id: gridSecurity.py,v 1.1.4.23 2006/03/27 09:13:39 harakaly Exp $
#####################################################################################

import os,string,socket,re,pwd,grp,time,stat
import sys, posix
from  gLiteInstallerLib import gLib
import urllib

class MyURLopener(urllib.FancyURLopener):
    def http_error_default(self, url, fp, errcode, errmsg, headers):
        void = fp.read()
        fp.close()
        raise IOError, ('http error', errcode, errmsg, headers)
        
urllib._urlopener = MyURLopener()

class gridSecurity:

    def __init__(self, glib=None):
        self.verbose = 0
        self.params = {}
        self.hostname = os.environ["HOSTNAME"].split('.')[0]
        self.fq_hostname = os.environ["HOSTNAME"]

        if glib == None:
            self.glib = gLib()
        else:
            self.glib = glib
        
        self.glib.verbose = self.verbose
        
    def setVerbosity(self,verbose):
        self.verbose = verbose
        self.glib.verbose = self.verbose
        
    #================================================================================
    # create_pool_accounts: function to manage the pool account creation
    #    Parameters:
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def create_pool_accounts(self):

        print "\nCreating/Verifying the pool accounts"

        #-------------------------------------------------------------------------------
        # Create N mapped accounts for each VO after having added the 'mapped' group
        #-------------------------------------------------------------------------------
    
        gridmap_dir = self.params['host.gridmap.dir']
        if not os.path.exists(gridmap_dir):
            os.makedirs(gridmap_dir)

        if not self.params.has_key('account.discovery'):
            self.params['account.discovery'] = 'false'
            
        volist = self.glib.getVoList()
        
        for vo in volist:
           
            voparams = self.glib.getVoParams(vo)

            if self.params['account.discovery'] == 'false':
            # If account discovery is false, check pool parameters,
            # create groups and accounts, set environment and verify interactivity
                
                if voparams.has_key('pool.account.group') and \
                    voparams.has_key('pool.account.basename') and \
                    voparams.has_key('pool.account.number'):
                        
                    pgroup = voparams['pool.account.group']
                    pbase = voparams['pool.account.basename']
                    pnumber = voparams['pool.account.number']
                
                    if voparams.has_key('pool.lsfgid'):
                        if voparams['pool.lsfgid'] == "":
                            lsfgid = '0'
                        else:
                            lsfgid = voparams['pool.lsfgid']
                    else:
                        lsfgid = '0'
                        
                    self.glib.printInfoMessage("Pool accounts for VO: %s" % voparams['vo.name'])
    
                    # Create group if doesn't exist, scripts terminates
                    # if lsfgid is specified and group already exist with different GID
                    if pgroup == "":
                        pgroup = pbase
                    if self.verbose:
                        print "==> Verifying group %s " % pgroup
                    if int(lsfgid):
                        gid = self.glib.add_group(pgroup,lsfgid)
                    else:
                        gid = self.glib.add_group(pgroup)
                    if self.verbose:
                        self.glib.printInfoMessage("Group %s has been added or verified" % pgroup)
            
                    # Create pool accounts in given group, this also sets the environment for the users
                    self.create_pool(pbase,int(pnumber),pgroup)
                    for acc in xrange(1,int(pnumber)+1):
                        if self.glib.checkInteractiveLogin("%s%03d" % (pbase,acc)):
                            return 1
            else:
            # Try to discover the existing accounts using base name and numeric suffix pattern
                
                if voparams.has_key('pool.account.basename'):
        
                    # Locate pool accounts 
                    l = pwd.getpwall()
        
                    self.glib.printInfoMessage("Discovering pool accounts for VO: %s" % voparams['vo.name'])
                    discoacc = []
                    for acc in l:
                        if re.match('%s[0-9]+$' % voparams['pool.account.basename'],acc[0]):
                            discoacc.append(acc[0])
                            
                    for acc in discoacc:
                        self.glib.setUserEnv(acc)
                        # Touch the gridmap mapping files if they don't exist
                        if not os.path.exists("%s/%s" % (gridmap_dir,acc)):
                            os.system("/bin/touch %s/%s" % (gridmap_dir,acc))
                        if self.glib.checkInteractiveLogin("%s" % acc):
                            return 1
    
        print "Creating/Verifying the pool accounts                     ",
        self.glib.printOkMessage()

        return 0
        
    #================================================================================
    # create_pool: function to create a set of pool account given a basename and number
    #    Parameters:
    #        basename: the pool base name
    #        number: the number of pool accounts to create
    #        group: an optional group to which the accounts should belong, if omitted
    #               account will belong to a Red Hat-style group
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def create_pool(self,basename,number, group=''):
        
        count = 1
        
        while count <= number:
        
            user = basename + "%03d" % count
            if group == '':
                group = user
            self.glib.add_user(user,group)
            count = count + 1
            if self.params.has_key('host.gridmap.dir'):
                gridmap_dir = self.params['host.gridmap.dir']
            else:
                gridmap_dir = '/etc/grid-security/gridmapdir'

            if not os.path.exists(gridmap_dir):
                os.makedirs(gridmap_dir)
    
            # Touch the gridmap mapping files if they don't exist
            if not os.path.exists("%s/%s" % (gridmap_dir,user)):
                os.system("/bin/touch %s/%s" % (gridmap_dir,user))
                
        return 0
    
    #================================================================================
    # configure_mkgridmap: function to configure the mkgridmap script. The function
    #                      checks if the right parameters are present, so it can be
    #                      called without previous checks
    #    Parameters:
    #        [OPTIONAL] account: empty if the mapping accounts are pool account base names,
    #                            name of parameter contaning the account names if they are accounts
    #        [OPTIONAL] role: empty if the accounts must be mapped to the VO root,
    #                         the name of the parameter containing the VO role if the accounts
    #                         must be mapped to a specific role
    #    Return:
    #        0 = ok
    #        1 = fail
    #        2 = warning
    #================================================================================
    def configure_mkgridmap(self,account="",role=""):

        if not self.params.has_key('install.mkgridmap.cron'):
            return 0

        if self.params['install.mkgridmap.cron'] == "false":
            return 0
                
        mkgc_file = self.params['mkgridmap.conf']
        mkgc_file1 = '%s.1' % self.params['mkgridmap.conf']

        if self.verbose:
            print "Updating the %s file..." % mkgc_file

        if os.path.exists(mkgc_file1):
            os.remove(mkgc_file1)
        if os.path.exists(mkgc_file):
            os.rename(mkgc_file, mkgc_file1)

        os.system("/bin/touch %s" % mkgc_file)

        volist = self.glib.getVoList()
        for vo in volist:
   
            voparams = self.glib.getVoParams(vo)
            voparams.update(self.params)

            if account == "" and not voparams.has_key('pool.account.basename'):
                self.glib.printWarningMessage("No pool accounts specified, skipping grid-mapfile configuration")
                self.glib.printWarningMessage("Please configure the mkgridmap.conf file manually")
                return 2

            if account != "" and not voparams.has_key(account):
                self.glib.printErrorMessage('The parameter %s is not defined in the configuration file' % account)
                return 1

            if role != "":
                if not voparams.has_key(role):
                    self.glib.printErrorMessage("The %s parameter is not defined" % role)
                    return 1
                
            vp1 = voparams['vo.name']
            vp2 = voparams['voms.hostname']
            if account == "":
                vp3 = voparams['pool.account.basename'].lower() 
            else:
                if not voparams.has_key(account):
                    self.glib.printErrorMessage("The %s parameter is not defined" % account)
                    return 1
                vp3 = voparams[account].lower() 
     
            if self.verbose:
                print "Creating\Verifying group entry for VO: %s..." % vp1

            container = "/%s" % vp1
            if role != "" and voparams[role] != "":
                container = container + "/Role=%s" % voparams[role]
    
            if account == "":
                groupentry = "group vomss://%s:8443/voms/%s?%s .%s" \
                    % (vp2, vp1, container, vp3)
            else:
                groupentry = "group vomss://%s:8443/voms/%s?%s %s" \
                    % (vp2, vp1, container, vp3)

            file = open(mkgc_file, 'r')
            for line in file:
                if line.find(groupentry) != -1:
                    break
            else:
                file.close()
                file = open(mkgc_file, 'a')
                file.write("\n" + groupentry + "\n")  

            file.close()

        if not os.path.exists(mkgc_file):
            print "Writing the %s file                              " % mkgc_file
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Failed to create %s" % mkgc_file)
            return 1

        if self.verbose:
            print "Writing the %s file                              " % mkgc_file,
            self.glib.printOkMessage()

        return 0

    #================================================================================
    # addFQAN2Gridmapfile: function to add properly formatted VOMS FQAN strings
    #                      mappings to pool accounts 
    #    Parameters:
    #        [OPTIONAL] account: empty if the mapping accounts are pool account base names,
    #                            the parameter name containing the account names if they are accounts
    #        [OPTIONAL] role: empty if the accounts must be mapped to the VO root,
    #                         the name of parameter containing the VO role if the accounts
    #                         must be mapped to a specific role
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def addFQAN2Gridmapfile(self,account="",role=""):
        if self.glib.params.has_key('host.gridmapfile.update') and self.glib.params['host.gridmapfile.update'] == 'false':
            return 0
        if self.verbose:
            print "Updating the grid-mapfile %s..." % self.params['host.gridmapfile']
        
        # Touch grid-mapfile if it doesn't exist
        if not os.path.exists(self.params['host.gridmapfile']):
            os.system("/bin/touch %s" % self.params['host.gridmapfile'])

        volist = self.glib.getVoList()
        for vo in volist:
           
            voparams = self.glib.getVoParams(vo)

            if account != "":
                if not voparams.has_key(account):
                    self.glib.printErrorMessage("The %s parameter is not defined" % account)
                    return 1
                        
            if role != "":
                if not voparams.has_key(role):
                    self.glib.printErrorMessage("The %s parameter is not defined" % role)
                    return 1
                        
            if account == "":
                # Use pool accounts
    
                if not voparams.has_key('pool.account.basename'):
                    self.glib.printWarningMessage("No pool accounts specified, skipping grid-mapfile configuration")
                    return 0
    
                if self.verbose:
                    print "Adding pool account mappings to VO FQAN entries..."
        
                for entry in ['/%s .%s' % (voparams['vo.name'], voparams['pool.account.basename'].lower()),
                              '/%s/* .%s' % (voparams['vo.name'], voparams['pool.account.basename'].lower())]:
                    
                    if self.verbose:
                        print "Adding/Verifying '%s'" % entry
                    
                    file = open(self.params['host.gridmapfile'], 'r' )
                    for line in file.xreadlines():
                        if line.find('%s' % entry) != -1:
                            file.close()
                            break
                    else:
                        file.close()
                        file = open(self.params['host.gridmapfile'], 'a')
                        file.write('%s\n' % entry)
                        file.close()
        
            else:
                # Use named accounts
    
                if self.verbose:
                    print "Adding account mappings to VO FQAN entries..."
        
                roleString = ""
                if role != "" and voparams[role] != "":
                    if voparams[role] == "*":
                        roleString = "%s" % params[role]
                    else:
                        roleString = "Role=%s" % voparams[role]

                if roleString != "":
                    
                    entry  = '/%s/%s %s' % (voparams['vo.name'], roleString, voparams[account].lower())
                    
                    if self.verbose:
                        print "Adding/Verifying '%s'" % entry
                    
                    file = open(self.params['host.gridmapfile'], 'r' )
                    for line in file.xreadlines():
                        if line.find('%s' % entry) != -1:
                            file.close()
                            break
                    else:
                        file.close()
                        file = open(self.params['host.gridmapfile'], 'a')
                        file.write('%s\n' % entry)
                        file.close()

        return 0
        
    #================================================================================
    # createGroupmapfile: function create and update the groupmapfile 
    #    Parameters:
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def createGroupmapfile(self):

        groupmap_file = "%s" % self.params['host.groupmapfile']
    
        if self.verbose:
            print "Updating the group mapfile %s..." % groupmap_file
    
        if not os.path.exists(groupmap_file):
            os.system("/bin/touch %s" % groupmap_file)

        volist = self.glib.getVoList()
        for vo in volist:
           
            voparams = self.glib.getVoParams(vo)

            if voparams.has_key('pool.account.basename'):
                
                for entry in ['/%s %s' % (voparams['vo.name'], voparams['pool.account.group']),
                              '/%s/* %s' % (voparams['vo.name'], voparams['pool.account.group'])]:
                    
                    if self.verbose:
                        print "Adding/Verifying '%s'" % entry
                
                    file = open(groupmap_file, 'r' )
                    for line in file.xreadlines():
                        if line.find(entry) != -1:
                            file.close()
                            break
                    else:
                        file.close()
                        file = open(groupmap_file, 'a')
                        file.write('%s\n' % entry)
                        file.close()
                
                if self.verbose:
                    print "Updating the group mapfile %s                    " % groupmap_file,
                    self.glib.printOkMessage()
            
            else:
                self.glib.printWarningMessage("No pool accounts specified, skipping groupmapfile configuration")
                self.glib.printWarningMessage("Please configure the groupmapfile file manually")

        return 0

    #================================================================================
    # createVomsdir: function to create the vomsdir directory and retrieve
    #                VOMS certificates from a list of URL if given
    #    Parameters:
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def createVomsdir(self):
        
        voms_dir = "%s" % os.environ['X509_VOMS_DIR']
        if self.verbose:
            print "Creating %s..." % voms_dir
    
        if not os.path.exists(voms_dir):
            os.makedirs(voms_dir,0755)
            print "Creating %s                                             " % voms_dir,
            self.glib.printOkMessage()
        
        volist = self.glib.getVoList()
        for vo in volist:
           
            voparams = self.glib.getVoParams(vo)

            if voparams.has_key('voms.cert.url'):
    
                vourl = voparams['voms.cert.url']
                if not vourl == '':
                    try:
                        if os.path.exists("%s/%s.pem" %( voms_dir, voparams['voms.hostname'])):
                            os.rename("%s/%s.pem" %( voms_dir, voparams['voms.hostname']), \
                                "%s/%s.pem.old" %( voms_dir, voparams['voms.hostname']))
                        urllib.urlretrieve(vourl, "%s/%s.pem" %( voms_dir, voparams['voms.hostname']))
                    except Exception, msg:
                        if os.path.exists("%s/%s.pem.old" %( voms_dir, voparams['voms.hostname'])):
                            os.rename("%s/%s.pem.old" %( voms_dir, voparams['voms.hostname']), \
                                "%s/%s.pem" %( voms_dir, voparams['voms.hostname']))
                        print "\n\n"
                        self.glib.printWarningMessage("Could not download VOMS server certificate from")
                        self.glib.printWarningMessage(vourl)
                        self.glib.printWarningMessage("The error was: %s" % msg)
                        self.glib.printWarningMessage("Please remember to copy the certificate manually to")
                        self.glib.printWarningMessage("%s before starting the services" % voms_dir)
                        print "\n\n"

                if self.verbose:
                    print "Dowloading VOMS certificate %s                       " % vourl,
                    self.glib.printOkMessage()
            
        return 0
        
    #================================================================================
    # createVomses: function to create and configure the vomses file
    #               containing VOMS Server properties
    #    Parameters:
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def createVomses(self):
        
        vomses_file = '%s/etc/vomses' % os.environ['GLITE_LOCATION']
        vomses_file1 = '%s/etc/vomses.1' % os.environ['GLITE_LOCATION']
    
        if self.verbose:
            print "Writing the %s file..." % vomses_file
    
        if os.path.exists(vomses_file1):
            os.remove(vomses_file1)
        if os.path.exists(vomses_file):
            os.rename(vomses_file,vomses_file1)

        file = open(vomses_file, 'w')
        
        volist = self.glib.getVoList()
        for vo in volist:
           
            voparams = self.glib.getVoParams(vo)

            vp1 = voparams['vo.name']

            vp2 = voparams['voms.hostname']

            if voparams['voms.cert.subj'] != "":
                vp3 = voparams['voms.cert.subj'] 
            elif os.path.exists("%s/%s.pem" % (os.environ['X509_VOMS_DIR'],voparams['voms.hostname'])):
                cert = "%s/%s.pem" % (os.environ['X509_VOMS_DIR'],voparams['voms.hostname'])
                vp3 = self.glib.get_certificate_subject(cert,sep='/')
            elif os.path.exists('/etc/grid-security/vomsdir/%s.pem' % voparams['voms.hostname']):
                cert = '/etc/grid-security/vomsdir/%s.pem' % voparams['voms.hostname']
                vp3 = self.glib.get_certificate_subject(cert,sep='/')
            elif os.path.exists('/etc/grid-security/vomsdir/%s/%s.pem' % (vp1,voparams['voms.hostname'])):
                cert = '/etc/grid-security/vomsdir/%s/%s.pem' % (vp1,voparams['voms.hostname'])
                vp3 = self.glib.get_certificate_subject(cert,sep='/')
            else:
                self.glib.printWarningMessage('It was not possible to find the certificate subject for VO %s' % vo)
                self.glib.printWarningMessage('in any of the standard locations')
                self.glib.printWarningMessage('Please set the value of the vomses entry')
                self.glib.printWarningMessage('in the file %s' % vomses_file)
                self.glib.printWarningMessage('before starting the services')
                vp3 = ""

            vp4 = voparams['voms.port.number'] 
     
            if self.verbose:
                print "Creating vomses entry for VO: %s..." % voparams['vo.name']
            
            file.write(' "%s" "%s" "%s" "%s" "%s"\n' % (vp1, vp2, vp4, vp3, vp1 ))  
    
        file.close()
        
        if not os.path.exists(vomses_file):
            print "Writing the %s file                          " % vomses_file
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Failed to create\\update %s" % vomses_file)
            if os.path.exists(vomses_file1):
                self.glib.printErrorMessage("Restoring previous file")
                os.rename(vomses_file1,vomses_file)
            return 1
    
        if self.verbose:
            print "Writing the %s file                          " % vomses_file,
            self.glib.printOkMessage()
    
        return 0

    #================================================================================
    # addSgmAccounts: function to create and configure the <vo>sgm accounts
    #                 used to perform special provider operations. The VO name
    #                 is converted to lower case because of restrictions on Linux
    #    Parameters:
    #    Return:
    #        0 = ok
    #        1 = fail
    #================================================================================
    def addSgmAccounts(self):
        
        if not self.params.has_key('create.sgm.account'):
            return 0
            
        if self.params['create.sgm.account'] == 'false':
            return 0

        volist = self.glib.getVoList()
        for vo in volist:
           
            voparams = self.glib.getVoParams(vo)

            if self.verbose:
                print "Creating sgm accounts and directories..." 
    
            if not voparams.has_key('vo.sgm.user'):
                voparams['vo.sgm.user'] = vo.lower() + "sgm"
                    
            if not voparams.has_key('vo.sgm.group'):
                voparams['vo.sgm.group'] = vo.lower() + "sgm"
    
            if not voparams.has_key('vo.sgm.vo.role'):
                self.glib.printErrorMessage('The parameter vo.sgm.vo.role is not defined in the VO List file')
                return 1
    
            # Linux user and groups can only be lower case
            voparams['vo.sgm.user'] = voparams['vo.sgm.user'].lower()
            voparams['vo.sgm.group'] = voparams['vo.sgm.group'].lower()

            VOTopDir = "%s/var/info" % os.environ['GLITE_LOCATION']
            if not os.path.exists(VOTopDir):
                os.makedirs("%s/var/info" % os.environ['GLITE_LOCATION'])
    
            voName = voparams['vo.name']
            voUser = voparams['vo.sgm.user']
            voGroup = voparams['vo.sgm.group']
            VODir  = VOTopDir + "/" + voName
            
            if self.verbose:
                print "Creating account %s" % voUser
        
            uid, gid = self.glib.add_user(voUser,voGroup)

            if uid == -1:
                print "Creating sgm accounts                                        ",
                self.glib.printFailedMessage()
                return 1

            if self.verbose:
                print "Creating VO directory %s" % VODir
        
            self.glib.check_dir(VODir,0770,uid,gid)

            if self.verbose:
                print "Updating grid-mapfile"
        
            self.addFQAN2Gridmapfile("vo.sgm.user","vo.sgm.vo.role")

            if self.verbose:
                print "Updating mkgridmap.conf file"
        
            self.addFQAN2Gridmapfile("vo.sgm.user","vo.sgm.vo.role")

            # Update glite-mkgridmap file for sgm accounts
            if self.configure_mkgridmap('vo.sgm.user','vo.sgm.vo.role'):
                print "Creating sgm accounts                                        ",
                self.glib.printFailedMessage()
                return 1
    
        print "Created sgm accounts and VO directories                      ",
        self.glib.printOkMessage()
    
        return 0
        
        
