#!/usr/bin/env python
#####################################################################################
#
# Script for configuring the gLite MYSQL service
#
# David Collados <david.collados@cern.ch>
# Joachim Flammer <Joachim.Flammer@cern.ch>
#
# Version 1.6.0
#
#####################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
#####################################################################################

import os
import time
import commands
import ConfigParser

from gLiteInstallerLib import gLib

class Mysql:

    #################################################################
    # start
    #
    # Start Mysql daemon
    #
    # return 0 ok
    #        1 error
    #     
    #################################################################
    def start(self):

        print "Starting MySQL daemon                                     ",
        errorcode = os.system("/etc/init.d/mysqld start")
        time.sleep(5)
        if errorcode:
            self.glib.printFailedMessage()
            return errorcode
        else:
            self.glib.printOkMessage()
            return 0    
    
    #################################################################
    # stop
    #
    # Stop Mysql daemon
    #
    # return 0 ok
    #        1 error
    #     
    #################################################################
    def stop(self):
        print "Stopping MySQL daemon                                     ",
        errorcode = os.system("/etc/init.d/mysqld stop")
        time.sleep(5)
        if errorcode:
            self.glib.printFailedMessage()
            return errorcode
        else:
            self.glib.printOkMessage()    
            return 0 
       
        
    #################################################################
    # init
    #
    # init Mysql daemon
    #
    # return 0 ok
    #        1 error
    #     
    #################################################################
    def init(self):
        return 0
        
        
    #################################################################
    # init
    #
    # init Mysql daemon
    #
    # return 0 ok
    #        1 error
    #     
    #################################################################
    def __init__(self):
        self.glib = gLib()
        return


    #################################################################
    # existsDB
    #
    # Checks whether a database exists or not. The mysqld daemon must be running    
    #
    # return 0 DB exist
    #        1 DB does not exists
    #     
    #################################################################
    def existsDB(self, database_name, mysql_root_password=""):
        retVal = self.existsDBWithPwd(database_name, mysql_root_password)
        return retVal


    #################################################################
    # existsDB
    #
    # Checks whether a database exists or not using mysql root password. 
    # The mysqld daemon must be running
    #
    # return 0 DB exist
    #        1 DB does not exists
    #     
    #################################################################
    def existsDBWithPwd(self, database_name, mysql_root_password):
    
        if mysql_root_password != "":
            if os.system('mysqlshow --password="%s" | grep " %s " > /dev/null 2>&1' % (mysql_root_password, database_name)) == 0:
                return 0
            else:
                return 1
        else:
            if os.system('mysqlshow | grep " %s " > /dev/null 2>&1' % database_name) == 0:
                return 0
            else:
                return 1

    
    #################################################################
    # checkPassword
    #
    # Check if database requires root password    
    #
    # return 0 no 
    #        1 yes
    #     
    #################################################################    
    def checkPassword(self, hostname="localhost"):

        print "Verifying if MySQL DB is password-protected..."

        # evaluate if password is set
        errorcode = os.system('mysql -u root --exec=quit')
        if errorcode:
            print "Root password is set"
            return 1
        else:
            print "Root password is not set"
            return 0


    #################################################################
    # checkPasswordSilently
    #
    # Check if database password matches silently without output   
    #
    # return  0: password matches
    #         1: password does not match
    #
    #################################################################    
    def checkPasswordSilently(self, password):

        if (password == ""):
            # check empty password
            status_output = commands.getstatusoutput("mysql -u root --exec=quit")
            status = status_output[0]/256
            output = status_output[1]
        else:
            # check with password
            status_output = commands.getstatusoutput('mysql --password="%s" --exec=quit' % password)
            status = status_output[0]/256
            output = status_output[1]
        
        if status !=0:
            return 1
        else:
            return 0

    #################################################################
    # checkMySQLConfiguration
    #
    # Check MySQL password configuration, fail if no match.
    # If the given root password is not empty and the existing root
    # password is empty, then the password is set if setempty=1
    # otherwise an error message is output and the function fails (default)
    #
    # return 0 Ok
    #        1 Error
    #
    #################################################################
    def checkMySQLConfiguration(self,rootpwd,setempty=0):

        if self.checkPasswordSilently(rootpwd) == 1:
            # password does not match
            if rootpwd != "":
                if self.checkPasswordSilently("") == 0:
                    if not setempty:
                        self.glib.printErrorMessage("Configuration detected unprotected DB (no password)")
                        self.glib.printErrorMessage("and a non-empty password specified in the configuration parameters.")
                        self.glib.printErrorMessage("Please set the password in the MySQL DB via:")
                        self.glib.printErrorMessage("  /usr/bin/mysqladmin -u root password 'yourPassword'")
                        self.glib.printErrorMessage("  /usr/bin/mysqladmin -u root -h %s password 'yourPassword'" % self.glib.fq_hostname)
                        self.glib.printErrorMessage("and/or correct the value in the configuration.")
                        return 1
                    else:
                        self.setpassword(rootpwd, self.glib.fq_hostname)
                else:
                    self.glib.printErrorMessage("MySQL password incorrect. Please check the specified password in the configuration.")
                    return 1
        return 0

    #################################################################
    # setpassword
    #
    # configure Database root password    
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################    
    def setpassword(self, password, hostname="localhost"):

        print "Trying to set root password for MySQL DB ..."

        # evaluate if password is already set (and if yes if correct password is given)
        errorcode = os.system('mysql -u root --exec=quit')
        if errorcode:
            self.glib.printWarningMessage("Root password already set. Checking if given root password is correct.")
            errorcode2 = os.system("mysql -u root --password='%s' --exec=quit" % password)
            if errorcode2:
                self.glib.printErrorMessage("Root password incorrect.")
                print "Setting root password for MySQL DB                        ",
                self.glib.printFailedMessage()
                return errorcode2
            else:
                self.glib.printInfoMessage("Root password correct. Previous error can be ignored.")
                print "Setting root password for MySQL DB                        ",
                self.glib.printOkMessage()            
                return 0
        else:
            print "Setting root password for MySQL DB                        ",
            errorcode3 =  os.system("/usr/bin/mysqladmin -u root password '%s'" % password)
            if hostname != 'hostname':
                errorcode3 += os.system("/usr/bin/mysqladmin -u root -h '%s' password '%s'" % (hostname, password))
            if errorcode3:
                self.glib.printFailedMessage()
                return errorcode3
            else:
                self.glib.printOkMessage()
                return 0

    #################################################################
    # create_db
    #
    # create Database
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################    
    def create_db(self, database_name, mysql_root_password=""):
        retVal = self.create_dbWithPwd(database_name, mysql_root_password)
        return retVal
 
 
     #################################################################
    # create_db
    #
    # create Database using root mysql root password
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################    
    def create_dbWithPwd(self, database_name, mysql_root_password):
        
        print "Creating MySQL database '%s'                     " % database_name,
        
        if mysql_root_password != "":
            errorcode = os.system('mysql -u root --password="%s" -e "CREATE DATABASE %s;"' % (mysql_root_password, database_name))
        else:
            errorcode = os.system('mysql -u root -e "CREATE DATABASE %s;"' % database_name)

        if errorcode:
            self.glib.printFailedMessage()
            return 1
            
        self.glib.printOkMessage()
        return 0 
                                                                                                                                                        

    #################################################################
    # add_user
    #
    # add user to DB
    #     
    #################################################################    
    def add_user(self, database_name, user_name, user_password, mysql_root_password=""):
        retVal = self.add_userWithPwd(database_name, user_name, user_password, mysql_root_password)
        return retVal
                                                                                                                                                       
    #################################################################
    # add_user
    #
    # add user to DB using root password
    #     
    #################################################################    
    def add_userWithPwd(self, database_name, user_name, user_password, mysql_root_password):
        
        # verify if DB exists
        if self.existsDBWithPwd(database_name,mysql_root_password) == 1:            
        
            # db does not exist - create
            errorcode = self.create_dbWithPwd(database_name,mysql_root_password)
            if errorcode:
                # creation not succesful
                return 1                    
                                       
        # add user
        print "Granting privileges for user '%s' on database '%s'                  " % (user_name, database_name),
    
        errorcode=0
        
        for host in ("","@\"%s\"" % os.environ['HOSTNAME'],"@localhost"):
            if mysql_root_password != "":
                errorcode += os.system("mysql -u root --password='%s' -e 'GRANT ALL PRIVILEGES on %s.* to \"%s\"%s IDENTIFIED BY \"%s\" WITH GRANT OPTION;'" % \
                    (mysql_root_password, database_name, user_name, host, user_password))
                errorcode += os.system("mysql -u root --password='%s' -e 'flush privileges;'" % mysql_root_password)
            else:
                errorcode += os.system("mysql -u root -e 'GRANT ALL PRIVILEGES on %s.* to %s%s IDENTIFIED BY \"%s\" WITH GRANT OPTION;'" % \
                    (database_name, user_name, host, user_password))
                errorcode += os.system("mysql -u root -e 'flush privileges;'")
        
        if errorcode:
            self.glib.printFailedMessage()
            return 1
    
        self.glib.printOkMessage()
        return 0
		
		
    #################################################################
    # add_user_old_password_format
    #
    # add user to DB with old password format
    #     
    #################################################################    
    def add_user_old_format(self, database_name, user_name, user_password, mysql_root_password=""):
        retVal = self.add_userWithPwd_Old_Format(database_name, user_name, user_password, mysql_root_password)
        return retVal
                                                                                                                                                       
    #################################################################
    # add_user with old password format
    #
    # add user to DB using root password with old password format
    #     
    #################################################################    
    def add_userWithPwd_Old_Format(self, database_name, user_name, user_password, mysql_root_password):
        # verify if DB exists
        if self.existsDBWithPwd(database_name,mysql_root_password) == 1:            
        
            # db does not exist - create
            errorcode = self.create_dbWithPwd(database_name,mysql_root_password)
            if errorcode:
                # creation not succesful
                return 1                    
                                       
        # add user
        print "Adding user '%s' to database '%s'                  " % (user_name, database_name),
    
        errorcode=0
        
        for host in ("","@\"%s\"" % os.environ['HOSTNAME'],"@localhost"):
            if mysql_root_password != "":
                errorcode += os.system("mysql -u root --password='%s' -e 'grant ALL PRIVILEGES on %s.* to \"%s\"%s identified by \"%s\";'" % \
                    (mysql_root_password, database_name, user_name, host, user_password))
                errorcode += os.system("mysql -u root --password='%s' -e 'set password for %s%s = OLD_PASSWORD(\"%s\");'" % (mysql_root_password,user_name,host,user_password))    
                errorcode += os.system("mysql -u root --password='%s' -e 'flush privileges;'" % mysql_root_password)

            else:
                errorcode += os.system("mysql -u root -e 'grant ALL PRIVILEGES on %s.* to %s%s identified by \"%s\";'" % \
                    (database_name, user_name, host, user_password))
                errorcode += os.system("mysql -u root -e 'set password for %s%s = OLD_PASSWORD(\"%s\");'" % (user_name, host,user_password))    
                errorcode += os.system("mysql -u root -e 'flush privileges;'")
        
        if errorcode:
            self.glib.printFailedMessage()
            return 1
    
        self.glib.printOkMessage()
        return 0
        
        
    #################################################################
    # restart
    #
    # restart mysql
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################    
    def restart(self):
        self.stop()
        self.start()
        return 0


    #################################################################
    # configure
    #
    # configure
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################    
    def configure(self):
        return 0    
 
    
    #################################################################
    # setConfiguration
    #
    # sets configuration in mysql configuration file
    #
    # Configuration file uses ini format with [section] section tags
    #
    # param section  section name for key value pair
    # param key      key entry to be set
    # param value    value entry to be set
    #
    # return 0 ok 
    #        1 error
    #     
    #################################################################            
    def setConfiguration(self, section, key, value):
                        
        configFilePath = "/etc/my.cnf"
        config = ConfigParser.ConfigParser()

        if os.path.exists(configFilePath):
            configFileOld = file(configFilePath, "r")
            config.readfp(configFileOld)
        
        # add section if necessary
        if not config.has_section(section):
            config.add_section(section)
                
        # add key value - pair
        config.set(section,key,value)
        
        # write file
        configFileNew = open("%s.tmp" % configFilePath, "w")
        config.write(configFileNew)
        configFileNew.close()
            
        os.rename("%s.tmp" % configFilePath, configFilePath)
                
        return 0


    #################################################################
    # checkConfiguration
    #
    # checks if configuration is set in mysql configuration file
    #
    # Configuration file uses ini format with [section] section tags
    #
    # param section  section name for key value pair
    # param key      key entry to be set
    # param value    value entry to be set
    #
    # return 0 Found matching configuration
    #        1 Found configuration entry with different setting
    #        2 Did not find configuration entry
    #     
    #################################################################            
    def checkConfiguration(self, section, key, value):
                
        configFilePath = "/etc/my.cnf"
        config = ConfigParser.ConfigParser()

        # check if configuration file exists
        if os.path.exists(configFilePath):
            configFileOld = file(configFilePath, "r")
            config.readfp(configFileOld)
        else:
            # file not present
            return 2
        
        # check if section exists
        if not config.has_section(section):
            # section not present
            return 2
        
        # check if option exists
        if not config.has_option(section,key):
            # option not present in section
            return 2
                
        # check if value is the same
        if not config.get(section, key) == value:
            # different setting
            return 1
                        
        return 0
          
            