#!/usr/bin/env python
#####################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
#####################################################################################
# stork v. 0.1.0
#
# Post-installation script for stork
# Robert Harakaly  < robert.harakaly@cern.ch >
#
# Version info: $Id: stork.py,v 1.5 2005/01/22 21:02:51 harakaly Exp $
#
# Usage: python stork [-c|-v|-h|--help]
#        -c          print configuration
#        -v          print version
#        -h,--help   print usage info
#
# Return codes: 0 - Ok
#               1 - Configuration failed
#
#####################################################################################
import os, types
from  gLiteInstallerLib import gLib

class Stork:
    def __init__(self, params):
        self.params = params
        self.glib = gLib()
        
    def start(self):
        print "Starting Stork"
        condormaster = '%s/sbin/condor_master' % self.params["STORK_LOCATION"]
        cmd = '/bin/su - %s -c "export CONDOR_CONFIG=%s; %s"' % (self.params['STORK_USER'], \
            self.params["CONDOR_CONFIG"], condormaster)
        os.system(cmd)
        
    def stop(self):
        pid = self.glib.getPID('condor_master')
        if pid != 0:
            os.system("kill -QUIT %d" % pid)
            print "The Condor master has been terminated"
        else:
            if self.verbose:
                print "The Condor master was not running"

    def configure(self):
        self.glib.add_user(self.params['stork.user.name'])
        self.modifyCondorConfigFiles()
        
    def modifyCondorConfigFiles(self):

        ##########################################
        # Modify global condor config file
        ##########################################
        condor_config = "%s/etc/condor_config" % self.params['stork.location']
        condor_config_orig = "%s.orig" % condor_config
        (uid,gid) = self.glib.get_user_info(self.params['stork.user.name'])
        
        if not os.path.exists(condor_config):
            print "Condor configuration file %s does not exist" % condor_config
            return 1

        os.rename(condor_config, condor_config_orig )

        try:
            fo = open(condor_config_orig,'r')
        except:
            print "Failed to save backup copy of Condor configuration file." % condor_config_orig
            return 1

        try:
            fn = open(condor_config,'w')
        except:
            print "%s could not be opened for writing." % condor_config
            return 1

        for l in fo.xreadlines():
            l = replaceToken(l,"stork.location","RELEASE_DIR = %s", self.params)
            l = replaceToken(l,"stork.local.dir","LOCAL_DIR = %s", self.params)
            l = replaceToken(l,"stork.local.configfile","LOCAL_CONFIG_FILE = %s", \
                {"stork.local.configfile":"%s/condor_config.local" % self.params['stork.local.dir']})
            l = replaceToken(l,"stork.admin","CONDOR_ADMIN = %s", self.params)
            l = replaceToken(l,"uid.domain","UID_DOMAIN = %s", {"uid.domain":"$(FULL_HOSTNAME)"})
            l = replaceToken(l,"filesystem.domain","FILESYSTEM_DOMAIN = %s", {"filesystem.domain":"$(FULL_HOSTNAME)"})
            l = replaceToken(l,"stork.credd.args","CREDD_ARGS = %s", self.params)
            l = replaceToken(l, ("uid","gid"), "CONDOR_IDS = %s.%s", {"uid":uid ,"gid":gid})            
            fn.write(l)

        fn.close()
        fo.close()
        
        #############################################
        #    Create local condor dirctory
        ############################################
        self.glib.check_dir(self.params['stork.local.dir'])
        self.glib.check_dir("%s/log" % self.params['stork.local.dir'],0,uid,gid)
        self.glib.check_dir("%s/cred_dir" % self.params['stork.local.dir'],0,uid,gid)
        self.glib.check_dir("%s/execute" % self.params['stork.local.dir'],0,uid,gid)
        self.glib.check_dir("%s/spool" % self.params['stork.local.dir'],0,uid,gid)
        ############################################
        # Modify local condor config file
        ############################################
        
        condor_config_local = '%s/condor_config.local' % self.params['stork.local.dir']
        condor_config_local_orig = '%s/local.%s/condor_config.local' % (self.params['stork.location'], self.glib.hostname)

        try:
            fo = open(condor_config_local_orig,'r')
        except:
            print "\tFailed to save backup copy of Condor local configuration file." % condor_config_local_orig
            return 1

        try:
            fn = open(condor_config_local,'w')
        except:
            print "%s could not be opened for writing." % condor_config_local
            return 1
                    
        for l in fo.xreadlines():
            l = replaceToken(l,"stork.location","RELEASE_DIR = %s", self.params)
            l = replaceToken(l,"stork.local.dir","LOCAL_DIR = %s", self.params)
            l = replaceToken(l,"stork.local.configfile","LOCAL_CONFIG_FILE = %s", \
                {"stork.local.configfile":"%s/condor_config.local" % self.params['stork.local.dir']})
            l = replaceToken(l,"stork.admin","CONDOR_ADMIN = %s", self.params)
            l = replaceToken(l,"uid.domain","UID_DOMAIN = %s", {"uid.domain":"$(FULL_HOSTNAME)"})
            l = replaceToken(l,"filesystem.domain","FILESYSTEM_DOMAIN = %s", {"filesystem.domain":"$(FULL_HOSTNAME)"})
            l = replaceToken(l,"stork.credd.args","CREDD_ARGS = %s", self.params)
            l = replaceToken(l, ("uid","gid"), "CONDOR_IDS = %s.%s", {"uid":uid ,"gid":gid})
            fn.write(l)

        fn.close()
        fo.close()

    def modifyStorkConfig(self):
        pass
        
def replaceToken(str, name, format, param):
    if isinstance(name, types.TupleType):
        if param.has_key(name[0]) & param.has_key(name[1]):
            if str.strip().find(format.split(" ")[0]) == 0:
                str = format % (param[name[0]], param[name[1]]) + "\n"
                print "replaceToken: " + str
            
    else:
        if param.has_key(name):
            if str.strip().find(format.split(" ")[0]) == 0:
                str = format % param[name] + "\n"
                print "replaceToken: " + str
    return str