#!/usr/bin/env python
#####################################################################################
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://eu-egee.org/partners/ for details on the copyright holders.
# For license conditions see the license file or http://eu-egee.org/license.html
#
#####################################################################################
#
# Script for configuring the Tomcat service for gLite
# Robert Harakaly <robert.harakaly@cern.ch>
# Joachim Flammer <joachim.flammer@cern.ch>
# Version info: $Id: tomcat5.py,v 1.23.2.2.6.3.4.4 2007/11/08 08:06:36 dshiyach Exp $
#
#####################################################################################

import os,pwd,shutil,commands
import sys
##import _xmlplus
##import _xmlplus.xpath as xpath
from Ft.Xml import XPath as xpath
import xml.dom.minidom
from xml.dom.minidom import Node
#from xml.dom.ext import PrettyPrint
from Ft.Xml.Lib.Print import PrettyPrint
from Ft.Xml.Domlette import NonvalidatingReader
from gLiteInstallerLib import gLib
        
class Tomcat5:

    def __init__(self,params):
        #initialize gLib
        self.glib = gLib()
        self.glib.set_params(params)

        self.params = params
        self.verbose = 0
                  

    #########################################
    # start
    #
    # start tomcat
    #
    #########################################    
    def start(self):
        print "Starting Tomcat                                           ",    
        retval = os.system("/etc/init.d/tomcat5 restart")
        if retval != 0:
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Please verify Tomcat and rerun the script")
            return 1
            
        self.glib.printOkMessage()
        return 0


    #########################################
    # stop
    #
    # stop tomcat
    #
    #########################################    
    def stop(self):
        
        # check if tomcat is running
        tomcat_lock_file = "/var/lock/subsys/tomcat5"
        if os.path.exists(tomcat_lock_file):
            print "Stopping Tomcat                                           ",
            retval = os.system("/etc/init.d/tomcat5 stop")
            if retval != 0:
                 self.glib.printFailedMessage()
                 self.glib.printErrorMessage("Please verify Tomcat and rerun the script")
                 return 1
            else:
                self.glib.printOkMessage()
                return 0
        
        self.glib.printInfoMessage("Tomcat is not running.");            
        return 0

    #########################################
    # configure
    #
    # configure tomcat (non-secure)
    #
    #########################################    
    def configure(self):
        
        error_level = 0
        
        # Check for JAVA_HOME
        if not os.environ.has_key('CATALINA_HOME'):
            if os.path.exists('/var/lib/tomcat5'):
                self.glib.export('CATALINA_HOME', "/var/lib/tomcat5")
            else:
                self.glib.printErrorMessage("CATALINA_HOME is not defined and the default location '/var/lib/tomcat5' doesn't exist")
                self.glib.printErrorMessage("Configuration cannot continue")
                return 1
        else:
            self.glib.export('CATALINA_HOME')
            
        # Check for JAVA_HOME
        if self.glib.checkJava() != 0:
            return 1

        # Create symlink to /etc/tomcat5
        if os.path.exists("/var/lib/tomcat5") \
           and os.path.exists("/etc/tomcat5") \
           and not os.path.exists("/var/lib/tomcat5/conf") \
           and os.environ.has_key('CATALINA_HOME') \
           and os.environ['CATALINA_HOME'] == "/var/lib/tomcat5" :

            os.symlink("/etc/tomcat5", "/var/lib/tomcat5/conf")


        # Set JAVA_HOME value in tomcat5.conf file
        if not os.path.exists("/etc/tomcat5/tomcat5.conf"):
            self.glib.printWarningMessage("The standard '/etc/tomcat5/tomcat5.conf' configuration file cannot be found")
            self.glib.printErrorMessage("Please configure the JAVA_HOME entry manually if required")
            error_level = 2
        else:
            t5org = open("/etc/tomcat5/tomcat5.conf",'r')
            t5new = open("/etc/tomcat5/tomcat5.conf.new",'w')
            optfound = 0
            for line in t5org.xreadlines():
                if line.find('JAVA_HOME=') != -1:
                    t5new.write("JAVA_HOME=%s\n" % os.environ['JAVA_HOME'])
                    optfound = 1
                else:
                    t5new.write(line)
            t5new.close()
            t5org.close()
            os.rename("/etc/tomcat5/tomcat5.conf.new", "/etc/tomcat5/tomcat5.conf")
            
            if not optfound:
                t5org = open("/etc/tomcat5/tomcat5.conf",'a')
                t5org.write("JAVA_HOME=%s\n" % os.environ['JAVA_HOME'])
                t5org.close()
                
            # determine memorysize
            proclist = os.popen("free | /bin/grep ^Mem | /bin/awk '{print int($2/1024/2)}'")
            proc = proclist.readlines()
            for item in proc:
                half_memorysize = item.strip()
            proclist.close()
            
            t5org = open("/etc/tomcat5/tomcat5.conf",'r')
            t5new = open("/etc/tomcat5/tomcat5.conf.new",'w')
            optfound = 0
            for line in t5org.xreadlines():
                if line.find('CATALINA_OPTS=') != -1:
                    t5new.write('CATALINA_OPTS="-Xmx%sM -server"\n' % half_memorysize)
                    optfound = 1
                else:
                    t5new.write(line)
            t5new.close()
            t5org.close()
            os.rename("/etc/tomcat5/tomcat5.conf.new", "/etc/tomcat5/tomcat5.conf")

            if not optfound:
                t5org = open("/etc/tomcat5/tomcat5.conf",'a')
                t5org.write('\nCATALINA_OPTS="-Xmx%sM -server"\n' % half_memorysize)
                t5org.close()
                
            uid,gid = self.glib.get_user_info('tomcat4')
            os.chown("/etc/tomcat5/tomcat5.conf",uid,gid)

        return error_level

    #########################################
    # setmaxthreads - OBSOLETE METHOD
    # - please use setConnectorParameter instead!
    #
    # set maximum number of threads for port
    # 
    # port       port to be updated
    # maxthreads max number of threads to be set
    #
    #########################################    
    def setmaxthreads(self, port, maxthreads):

        inputfile = '%s/conf/server.xml' % os.environ['CATALINA_HOME']
        doc = xml.dom.minidom.parse(inputfile)
        nodes = xpath.Evaluate("/Server/Service/Connector[@port='%s']/@maxThreads" % port, doc.documentElement)        
        for node in nodes:
            node.nodeValue="%s" % maxthreads
        try:        
            outputfile = '%s/conf/server.xml' % os.environ['CATALINA_HOME']
            output = open(outputfile, "w")
            output.writelines(doc.toprettyxml())
            output.close()
        except IOError, ex:
            if ex.strerror == 'Permission denied':
                print "\nFile is not writable"
                return 1
            else:
                raise ex                                
                return 1
        
        return 0


    #########################################
    # setConnectorParameter
    #
    # set parameter for a connector
    # 
    # port             port to be updated
    # parameterName    name of parameter to be updated
    # parameterValue   value of parameter to be updated
    #
    #########################################    
    def setConnectorParameter(self, port, parameterName, parameterValue):

        # read file
        inputfile = '%s/conf/server.xml' % os.environ['CATALINA_HOME']
        doc = xml.dom.minidom.parse(inputfile)
 
        # get connector
        nodes = xpath.Evaluate("/Server/Service/Connector[@port='%s']" 
        	% port, doc.documentElement)

        for node in nodes:
            # set attribute of connector
            node.setAttribute(parameterName,parameterValue)

        # write new file
        try:        
            outputfile = '%s/conf/server.xml' % os.environ['CATALINA_HOME']
            output = open(outputfile, "w")
            output.writelines(doc.toprettyxml())
            output.close()
        except IOError, ex:
            if ex.strerror == 'Permission denied':
                print "\nFile is not writable"
                return 1
            else:
                raise ex                                
                return 1
        
        # clean file
        inputfile = '%s/conf/server.xml' % os.environ['CATALINA_HOME']
        outputfile = '%s/conf/server.xml.1' % os.environ['CATALINA_HOME']
        input = open(inputfile,'r')
        output = open(outputfile, 'w')
        for line in input:
            if line.strip() != "":
                output.write(line)

        output.close()
        input.close()
        os.rename('%s/conf/server.xml.1' % os.environ['CATALINA_HOME'], '%s/conf/server.xml' % os.environ['CATALINA_HOME'])        
        
        return 0                
  
        
    #########################################
    # deploy
    #
    # deploy service to Tomcat
    # 
    # deploypath  deployment path
    # docbase	  document base
    #
    ######################################### 
    def deploy(self, deploypath, docbase):
    
        # remove old existing unpacked code (otherwise code will not be updated)
    	if os.path.exists("%s/webapps/%s" % (os.environ['CATALINA_HOME'] , deploypath)):
        	
            print "Removing old installed web applications                   ",
    	
            try:
                shutil.rmtree("%s/webapps/%s" % (os.environ['CATALINA_HOME'] , deploypath))
            except IOError, ex:
                self.glib.printFailedMessage()
                raise ex								
                return 1
            self.glib.printOkMessage()
        	
        # create Configuration 'web.xml'-type  file
        print "Deploying web application                                 ",
        outputfile = '%s/conf/Catalina/localhost/%s.xml' % (os.environ['CATALINA_HOME'], deploypath)
        try:                                        
            output = open(outputfile, 'w')
            output.write("<?xml version='1.0' encoding='utf-8'?>\n")
            output.write('<Context docBase=\"%s" path="/%s">\n' \
              % (docbase, deploypath))
            output.write('</Context>\n')
            output.close()        
        except IOError, ex:
            if ex.strerror == 'Permission denied':
                self.glib.printFailedMessage()
                self.glib.printErrorMessage("\nFile %s is not writable" % outputfile)
                return 1
            else:
                self.glib.printFailedMessage()
                raise ex                                
                return 1
        self.glib.printOkMessage()        
        return 0    

    #########################################
    # restart
    #
    # restart Tomcat
    # 
    ######################################### 
    def restart(self):
        os.system("/etc/init.d/tomcat5 restart")
        
        return 0

#####################################
# secure tomcat class
# 
#####################################
class Secure_tomcat5 (Tomcat5):

    #########################################
    # configure
    #
    # configure tomcat with security
    #
    # return 0 ok
    #        1 error
    #
    #########################################	
    def configure(self):

        # -------------------------------------------------------------------------------
        # Check tomcat parameters
        # -------------------------------------------------------------------------------

        print "Checking configuration parameters for tomcat              ",

        # Create symlink to /etc/tomcat5
        if os.path.exists("/var/lib/tomcat5") \
           and os.path.exists("/etc/tomcat5") \
           and not os.path.exists("/var/lib/tomcat5/conf") \
           and os.environ.has_key('CATALINA_HOME') \
           and os.environ['CATALINA_HOME'] == "/var/lib/tomcat5" :

            os.symlink("/etc/tomcat5", "/var/lib/tomcat5/conf")

        # host certificate file			
        if self.params.has_key('host.certificate.file'):
            host_certificate_file = self.params['host.certificate.file']
        else:
            self.glib.printFailedMessage()
            printErrorMessage("Parameter \"host.certificate.file\" not found.")
            return 1

        # host.key.file
        if self.params.has_key('host.key.file'):
            host_key_file = self.params['host.key.file']
        else:
            self.glib.printFailedMessage()
            printErrorMessage("Parameter \"host.key.file\" not found.")
            return 1			

        # tomcat user name
        if self.params.has_key('tomcat.user.name'):
            tomcat_user_name = self.params['tomcat.user.name']
        else:
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Parameter \"tomcat.user.name\" not found.")
            return 1
        
        # tomcat user group
        if self.params.has_key('tomcat.user.group'):
            tomcat_user_group = self.params['tomcat.user.group']
        else:
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Parameter \"tomcat.user.group\" not found.")
            return 1

        # user certificates path
        if self.params.has_key('user.certificate.path'):
            user_certificates_path = self.params['user.certificate.path']
        else:
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Parameter \"user.certificate.path\" not found.")
            return 1
        
        # tomcat conf options
        if self.params.has_key('tomcat.CATALINA_OPTS'):
            catalinaOpts = self.params['tomcat.CATALINA_OPTS']
        else:
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Parameter \"tomcat.CATALINA_OPTS\" not found.")
            return 1

        self.glib.printOkMessage()

        # -------------------------------------------------------------------------------
        # Create tomcat hostkeys and certificates
        # -------------------------------------------------------------------------------

        print "Updating hostkeys and certificates for tomcat             ",
        
        tomcat_user_home_dir = pwd.getpwnam(tomcat_user_name)[5]

        user_cert_filename = "hostcert.pem"
        user_key_filename  = "hostkey.pem"

        # create tomcat user dir
        if not os.path.exists(tomcat_user_home_dir):
            os.mkdir("%s" % tomcat_user_home_dir,0755)
        
        # create dir user_certificates_path if it does not exist
        if not os.path.exists("%s/%s" % (tomcat_user_home_dir, user_certificates_path)):            
            os.mkdir("%s/%s" % (tomcat_user_home_dir, user_certificates_path),0755)
            os.system("chown %s:%s %s/%s" % (tomcat_user_name, tomcat_user_group, \
                       tomcat_user_home_dir, user_certificates_path))

        # copy key and cert file		
        errorcode = 0
        errorcode += os.system("cp %s %s/%s/%s" % \
            (host_key_file, tomcat_user_home_dir, user_certificates_path, user_key_filename))/256
        errorcode += os.system("cp %s %s/%s/%s" % \
            (host_certificate_file, tomcat_user_home_dir, user_certificates_path, user_cert_filename))/256
        if errorcode:
            self.glib.printFailedMessage()
            self.glib.printErrorMessage("Could not copy certificate and key for user '%s'" % tomcat_user_name)
            self.glib.printErrorMessage("to files '%s/%s/%s' and '%s/%s/%s'" % \
                    (tomcat_user_home_dir, user_certificates_path, user_cert_filename, \
                     tomcat_user_home_dir, user_certificates_path, user_key_filename))
            return 1
		
        # change owner of key and cert file		
        errorcode = 0
        errorcode += os.system("chown %s:%s %s/%s/%s" % \
             (tomcat_user_name, tomcat_user_group, \
              tomcat_user_home_dir, user_certificates_path, user_key_filename))/256
        errorcode += os.system("chown %s:%s %s/%s/%s" % \
             (tomcat_user_name, tomcat_user_group, \
              tomcat_user_home_dir, user_certificates_path, user_cert_filename))/256

        if errorcode:
             self.glib.printFailedMessage()
             self.glib.printErrorMessage("Could not change access rights of certificate and key for user '%s:%s'" % \
                (tomcat_user_name, tomcat_user_group))
             self.glib.printErrorMessage("for file '%s/%s/hostkey.pem' and '%s/%s/%s'" % \
                (tomcat_user_home_dir, user_certificates_path, user_key_filename, \
                 tomcat_user_home_dir, user_certificates_path, user_cert_filename))
             return 1

        self.glib.printOkMessage()    
    
        # -------------------------------------------------------------------------------
		# Update properties SSLKEY for security trustmanager and tomcat
        # -------------------------------------------------------------------------------
        
        print "Updating trustmanager configuration properties            ",
        
        inputFile  = open("%s/etc/glite-security-trustmanager/config.properties" % self.params['GLITE_LOCATION'], "r")
        outputFile = open("%s/etc/glite-security-trustmanager/config.properties~" % self.params['GLITE_LOCATION'], "w")

        for line in inputFile.readlines():
            if line.find('glite-security-trustmanager.SSLKEY') > -1:
                outputFile.write("glite-security-trustmanager.SSLKEY          = %s/%s/%s\n" % (tomcat_user_home_dir, user_certificates_path, user_key_filename))
            else:
                if line.find('glite-security-trustmanager.SSLCERTFILE') > -1:
                     outputFile.write("glite-security-trustmanager.SSLCERTFILE          = %s/%s/%s\n" % (tomcat_user_home_dir, user_certificates_path, user_cert_filename))
                else:
		             outputFile.write(line)
        inputFile.close()
        outputFile.close()
                		
        os.rename("%s/etc/glite-security-trustmanager/config.properties~" % self.params['GLITE_LOCATION'],\
                  "%s/etc/glite-security-trustmanager/config.properties" % self.params['GLITE_LOCATION'])        
		   
        self.glib.printOkMessage()
        
        # -------------------------------------------------------------------------------
        # run trustmanager configuration
        # -------------------------------------------------------------------------------

        if self.verbose:
            print "Configuring security trustmanager ...             "
    	else:
            print "Configuring security trustmanager                         ",

        errorcode = 0
    
        if self.verbose:
            errorcode = os.system("sh %s/etc/glite-security-trustmanager/configure.sh" % self.params['GLITE_LOCATION'])/256
        else:
            status_output = commands.getstatusoutput("sh %s/etc/glite-security-trustmanager/configure.sh" % self.params['GLITE_LOCATION'])
            errorcode = status_output[0]/256
            output = status_output[1]

        if errorcode == 4:
            self.glib.printInfoMessage("Reconfigured trustmanager")
        if self.verbose:
            print "Configuring security trustmanager                             ",
            self.glib.printOkMessage()
        else:
            if errorcode == 0:
                if self.verbose:
                    print "Configuring security trustmanager                         ",
                self.glib.printOkMessage()
            else:
                if self.verbose:
                    print "Configuring security trustmanager                         ",             
                self.glib.printFailedMessage()
                return 1  
        
        # Enable or disable unsecure port access
        if not self.params.has_key('allow.unsecure.port'):
            self.params['allow.unsecure.port'] = "false"
        
        if self.enableUnsecureConnector(self.params['allow.unsecure.port']) :
            return 1
            
        # -------------------------------------------------------------------------------
        # Check for CATALINA_HOME
        # -------------------------------------------------------------------------------

        if not os.environ.has_key('CATALINA_HOME'):
            if os.path.exists('/var/lib/tomcat5'):
                self.glib.export('CATALINA_HOME', "/var/lib/tomcat5")
            else:
                self.glib.printErrorMessage("CATALINA_HOME is not defined and the default location '/var/lib/tomcat5' doesn't exist")
                self.glib.printErrorMessage("Configuration cannot continue")
                return 1
        else:
            self.glib.export('CATALINA_HOME')

        # -------------------------------------------------------------------------------
        # Check for JAVA_HOME
        # -------------------------------------------------------------------------------
        if self.glib.checkJava() != 0:
            return 1

        # -------------------------------------------------------------------------------
        # Update tomcat global configuration options
        # -------------------------------------------------------------------------------

        # set JAVA_HOME
        if self.setTomcatConfEntry("JAVA_HOME", os.environ['JAVA_HOME']):
            return 1
                
        # set CATALINA_OPTS        
        if self.setTomcatConfEntry("CATALINA_OPTS", '"%s"' % catalinaOpts):
            return 1
        
        # -------------------------------------------------------------------------------
        # Set owner ship for tomcat configuration  
        # -------------------------------------------------------------------------------
                
        uid,gid = self.glib.get_user_info(tomcat_user_name)
        os.chown("/etc/tomcat5/tomcat5.conf",uid,gid)

        return 0


    #################################################
    # enableUnsecureConnector
    #
    # Enable/disable the unsecure connector in Tomcat
    # 
    # enable
    #        value: true | false
    #        type:  String
    #
    #################################################    
    def enableUnsecureConnector(self, enable="false"):

        inputfile = 'file:///%s/conf/server.xml' % os.environ['CATALINA_HOME']
  
        try :
            doc = NonvalidatingReader.parseUri (inputfile)
        except Exception, e:
            print "Error parsing Tomcat configuration file: %s" % e.message
            return 1
  
        nodes = xpath.Evaluate("/Server/Service[@name='Catalina']/Connector[@port='8080']", doc.documentElement)

        if enable == "true":
            if len(nodes) == 0:
                print "Enabling unsecure port 8080                               ",

                connector = doc.createElement('Connector')
                connector.setAttribute('port','8080')
                connector.setAttribute('minSpareThreads','25')
                connector.setAttribute('maxSpareThreads','75')
                connector.setAttribute('enableLookups','false')
                connector.setAttribute('acceptCount','100')
                connector.setAttribute('debug','0')
                connector.setAttribute('connectionTimeout','20000')
                connector.setAttribute('disableUploadTimeout','true')

                nodes = xpath.Evaluate("/Server/Service[@name='Catalina']", doc.documentElement)
                nodes[0].appendChild(connector)

                self.glib.printOkMessage()
        else:
            if len(nodes) > 0:
                print "Disabling unsecure port 8080         ",
                serviceNode = xpath.Evaluate("/Server/Service[@name='Catalina']", doc.documentElement)[0]
                for node in nodes:
                    serviceNode.removeChild(node)
                self.glib.printOkMessage()

        try:        
            outputfile = '%s/conf/server.xml' % os.environ['CATALINA_HOME']
#            output = open(outputfile, "w")
#            PrettyPrint(doc,stream=output)
#            output.close()
        except IOError, ex:
            if ex.strerror == 'Permission denied':
                print "\nFile is not writable"
                return 1
            else:
                return 1

        return 0       

        
    #########################################
    # conditionalStart
    #
    # conditional start: if server is already running
    # we restart it, otherwise we just start it
    #
    #########################################	
    def conditionalStart(self):
                
        if os.path.exists("/var/lock/subsys/tomcat5"):
             if os.path.exists("/var/run/tomcat5.pid"):
                self.glib.printInfoMessage("Tomcat is already running. Stopping and restarting Tomcat ...")
                errorcode = self.stop()
                if errorcode:
                   return 1
                   
        errorcode = self.start()
        if errorcode:
            return 1
        
        return 0
        
    
    #########################################
    # status
    #
    # get status of tomcat server
    #
    # return 0 running
    #        1 otherwise
    #
    #########################################    
    def status(self):
                
        if os.path.exists("/var/lock/subsys/tomcat5"):
             if os.path.exists("/var/run/tomcat5.pid"):
                return 0
        
        return 1
    
    
    #################################################
    # setTomcatConfEntry
    #
    # sets an entry in the tomcat5.conf file as key=value
    # 
    #        key:    key to be set
    #        value:  value to be set
    #
    #################################################    
    def setTomcatConfEntry(self, key, value):
        
        foundEntry = 0
        if os.path.exists('%s/conf/tomcat5.conf' % os.environ['CATALINA_HOME']):
            
            inputfile = open('%s/conf/tomcat5.conf' % os.environ['CATALINA_HOME'], 'r')
            outputfile = open('%s/conf/tomcat5.conf.tmp' % os.environ['CATALINA_HOME'], 'w')
                    
            # loop over input and write/replace line
            for line in inputfile.readlines():
                if line.startswith("%s=" % key):
                    outputfile.write("%s=%s\n" % (key, value))
                    foundEntry = 1
                else:
                    outputfile.write(line)
            if foundEntry==0:
                outputfile.write("%s=%s\n" % (key, value))
            
            inputfile.close()
            outputfile.close()
            
            # put file back to original place
            os.rename('%s/conf/tomcat5.conf.tmp' % os.environ['CATALINA_HOME'] , '%s/conf/tomcat5.conf' % os.environ['CATALINA_HOME'])
        
            return 0
        
        # file does not exist
        return 1
            
            
